from PySide6 import QtWidgets, QtCore, QtGui


class _OutsideNeumorphismEffect(QtWidgets.QGraphicsEffect):
    _cornerShift = (
        QtCore.Qt.TopLeftCorner, 
        QtCore.Qt.TopLeftCorner, 
        QtCore.Qt.BottomRightCorner, 
        QtCore.Qt.BottomLeftCorner
    )

    def __init__(self,
                 parent: object = None,
                 distance: int = 4, 
                 lightColor: QtGui.QColor = QtGui.QColor("#FFFFFF"),
                 darkColor: QtGui.QColor = QtGui.QColor("#7d7d7d"), 
                 clipRadius: int = 4,
                 origin: QtCore.Qt.Corner = QtCore.Qt.Corner.TopLeftCorner):
        super().__init__(parent)

        self._leftGradient = QtGui.QLinearGradient(1, 0, 0, 0)
        self._leftGradient.setCoordinateMode(QtGui.QGradient.CoordinateMode.ObjectBoundingMode)

        self._topGradient = QtGui.QLinearGradient(0, 1, 0, 0)
        self._topGradient.setCoordinateMode(QtGui.QGradient.CoordinateMode.ObjectBoundingMode)

        self._rightGradient = QtGui.QLinearGradient(0, 0, 1, 0)
        self._rightGradient.setCoordinateMode(QtGui.QGradient.CoordinateMode.ObjectBoundingMode)

        self._bottomGradient = QtGui.QLinearGradient(0, 0, 0, 1)
        self._bottomGradient.setCoordinateMode(QtGui.QGradient.CoordinateMode.ObjectBoundingMode)

        self._radial = QtGui.QRadialGradient(.5, .5, .5)
        self._radial.setCoordinateMode(QtGui.QGradient.CoordinateMode.ObjectBoundingMode)

        self._conical = QtGui.QConicalGradient(.5, .5, 0)
        self._conical.setCoordinateMode(QtGui.QGradient.CoordinateMode.ObjectBoundingMode)


        self._origin = origin

        distance = max(0, distance)
        self._clipRadius = min(distance, max(0, clipRadius))

        self._setColors(lightColor, darkColor)
        self._setDistance(distance)

    def setColors(self, 
                  color1 = QtGui.QColor("#FFFFFF"), 
                  color2 = QtGui.QColor("#7d7d7d")):
        # if (isinstance(color1, QtCore.Qt.GlobalColor) 
        #     and isinstance(color2, QtCore.Qt.GlobalColor)):

            # color1 = QtGui.QColor(color1)
            # color2 = QtGui.QColor(color2)

            self._setColors(color1, color2)
            self._setDistance(self._distance)

            self.update()

    def _setColors(self, color1, color2):
        self._baseStart = color1
        self._baseStop = QtGui.QColor(color1)
        self._baseStop.setAlpha(0)

        self._shadowStart = color2
        self._shadowStop = QtGui.QColor(color2)
        self._shadowStop.setAlpha(0)

        self.lightSideStops = [(0, self._baseStart), (1, self._baseStop)]
        self.shadowSideStops = [(0, self._shadowStart), (1, self._shadowStop)]
        self.cornerStops = [(0, self._shadowStart), (.25, self._shadowStop),
                            (.75, self._shadowStop), (1, self._shadowStart)]

        self._setOrigin(self._origin)

    def distance(self):
        return self._distance

    def setDistance(self, distance):
        if distance == self._distance:
            return

        self._setDistance(distance)
        self.updateBoundingRect()

    def _getCornerPixmap(self, rect, grad1, grad2=None):
        pm = QtGui.QPixmap(self._distance + self._clipRadius, 
                           self._distance + self._clipRadius)
        pm.fill(QtCore.Qt.GlobalColor.transparent)

        qp = QtGui.QPainter(pm)

        if self._clipRadius > 1:
            path = QtGui.QPainterPath()
            path.addRect(rect)
            size = self._clipRadius * 2 - 1
            mask = QtCore.QRectF(0, 0, size, size)
            mask.moveCenter(rect.center())
            path.addEllipse(mask)
            qp.setClipPath(path)
        qp.fillRect(rect, grad1)
        if grad2:
            qp.setCompositionMode(qp.CompositionMode.CompositionMode_SourceAtop)
            qp.fillRect(rect, grad2)
        qp.end()
        return pm

    def _setDistance(self, distance):
        distance = max(1, distance)
        self._distance = distance
        if self._clipRadius > distance:
            self._clipRadius = distance
        distance += self._clipRadius
        r = QtCore.QRectF(0, 0, distance * 2, distance * 2)

        lightSideStops = self.lightSideStops[:]
        shadowSideStops = self.shadowSideStops[:]

        if self._clipRadius:
            gradStart = self._clipRadius / (self._distance + self._clipRadius)
            lightSideStops[0] = (gradStart, lightSideStops[0][1])
            shadowSideStops[0] = (gradStart, shadowSideStops[0][1])

        # create the 4 corners as if the light source was top-left
        self._radial.setStops(lightSideStops)
        topLeft = self._getCornerPixmap(r, self._radial)

        self._conical.setAngle(359.9)
        self._conical.setStops(self.cornerStops)

        topRight = self._getCornerPixmap(r.translated(-distance, 0), 
                                         self._radial, self._conical)

        self._conical.setAngle(270)
        self._conical.setStops(self.cornerStops)
        bottomLeft = self._getCornerPixmap(r.translated(0, -distance), 
                                           self._radial, self._conical)

        self._radial.setStops(shadowSideStops)
        bottomRight = self._getCornerPixmap(r.translated(-distance, -distance), 
                                            self._radial)

        # rotate the images according to the actual light source
        images = topLeft, topRight, bottomRight, bottomLeft
        shift = self._cornerShift.index(self._origin)
        if shift:
            transform = QtGui.QTransform().rotate(shift * 90)
            for img in images:
                img.swap(img.transformed(transform, QtCore.Qt.SmoothTransformation))

        # and reorder them if required
        self.topLeft, self.topRight, self.bottomRight, self.bottomLeft = images[-shift:] + images[:-shift]

    def origin(self):
        return self._origin

    def setOrigin(self, origin):
        origin = QtCore.Qt.Corner(origin)
        if origin == self._origin:
            return

        self._setOrigin(origin)
        self._setDistance(self._distance)

        self.update()

    def _setOrigin(self, origin):
        self._origin = origin

        gradients = self._leftGradient, self._topGradient, self._rightGradient, self._bottomGradient
        stops = self.lightSideStops, self.lightSideStops, self.shadowSideStops, self.shadowSideStops

        # assign color stops to gradients based on the light source position
        shift = self._cornerShift.index(self._origin)
        for grad, stops in zip(gradients, stops[-shift:] + stops[:-shift]):
            grad.setStops(stops)

    def clipRadius(self):
        return self._clipRadius

    def setClipRadius(self, radius):
        if radius == self._clipRadius:
            return
        self._setClipRadius(radius)
        self.update()

    def _setClipRadius(self, radius):
        radius = min(self._distance, max(0, int(radius)))
        self._clipRadius = radius
        self._setDistance(self._distance)

    def boundingRectFor(self, rect):
        d = self._distance
        return rect.adjusted(-d, -d, d, d)

    def draw(self, qp):
        restoreTransform = qp.worldTransform()

        qp.setPen(QtCore.Qt.NoPen)
        x, y, width, height = self.sourceBoundingRect(QtCore.Qt.DeviceCoordinates).getRect()
        right = x + width
        bottom = y + height
        clip = self._clipRadius
        doubleClip = clip * 2

        if self._clipRadius:
            path = QtGui.QPainterPath()
            source = self.sourcePixmap(QtCore.Qt.DeviceCoordinates)
            sourceBoundingRect = self.sourceBoundingRect(QtCore.Qt.DeviceCoordinates)
            qp.save()

            qp.setTransform(QtGui.QTransform())
            path.addRoundedRect(sourceBoundingRect.x(), sourceBoundingRect.y(), sourceBoundingRect.width(),
                                sourceBoundingRect.height(), self._clipRadius, self._clipRadius)
            qp.setClipPath(path)
            qp.drawPixmap(sourceBoundingRect.x() - self._distance, sourceBoundingRect.y() - self._distance, source)
            qp.restore()
        else:
            path = QtGui.QPainterPath()
            source = self.sourcePixmap(QtCore.Qt.DeviceCoordinates)
            sourceBoundingRect = self.sourceBoundingRect(QtCore.Qt.DeviceCoordinates)
            qp.save()
            qp.setTransform(QtGui.QTransform())
            path.addRect(sourceBoundingRect.x(), sourceBoundingRect.y(), sourceBoundingRect.width(),
                         sourceBoundingRect.height())
            qp.setClipPath(path)
            qp.drawPixmap(sourceBoundingRect.x() - self._distance, sourceBoundingRect.y() - self._distance, source)
            qp.restore()

        qp.setWorldTransform(QtGui.QTransform())
        leftRect = QtCore.QRectF(x - self._distance, y + clip, self._distance, height - doubleClip)
        qp.setBrush(self._leftGradient)
        qp.drawRect(leftRect)

        topRect = QtCore.QRectF(x + clip, y - self._distance, width - doubleClip, self._distance)
        qp.setBrush(self._topGradient)
        qp.drawRect(topRect)

        rightRect = QtCore.QRectF(right, y + clip, self._distance, height - doubleClip)
        qp.setBrush(self._rightGradient)
        qp.drawRect(rightRect)

        bottomRect = QtCore.QRectF(x + clip, bottom, width - doubleClip, self._distance)
        qp.setBrush(self._bottomGradient)
        qp.drawRect(bottomRect)

        qp.drawPixmap(x - self._distance, y - self._distance, self.topLeft)
        qp.drawPixmap(right - clip, y - self._distance, self.topRight)
        qp.drawPixmap(right - clip, bottom - clip, self.bottomRight)
        qp.drawPixmap(x - self._distance, bottom - clip, self.bottomLeft)

        qp.setWorldTransform(restoreTransform)


class _InsideNeumorphismEffect(QtWidgets.QGraphicsEffect):
    _cornerShift = (
        QtCore.Qt.TopLeftCorner, 
        QtCore.Qt.TopRightCorner,
        QtCore.Qt.BottomRightCorner, 
        QtCore.Qt.BottomLeftCorner
    )

    def __init__(self, 
                 parent: object = None,
                 distance: int = 4, 
                 lightColor: QtGui.QColor = QtGui.QColor("#FFFFFF"),
                 darkColor: QtGui.QColor = QtGui.QColor("#7d7d7d"), 
                 clipRadius: int = 4,
                 origin: QtCore.Qt.Corner = QtCore.Qt.Corner.BottomRightCorner):
        super().__init__(parent)

        self._leftGradient = QtGui.QLinearGradient(0, 0, 1, 0)
        self._leftGradient.setCoordinateMode(QtGui.QGradient.ObjectBoundingMode)

        self._topGradient = QtGui.QLinearGradient(0, 0, 0, 1)
        self._topGradient.setCoordinateMode(QtGui.QGradient.ObjectBoundingMode)

        self._rightGradient = QtGui.QLinearGradient(1, 0, 0, 0)
        self._rightGradient.setCoordinateMode(QtGui.QGradient.ObjectBoundingMode)

        self._bottomGradient = QtGui.QLinearGradient(0, 1, 0, 0)
        self._bottomGradient.setCoordinateMode(QtGui.QGradient.ObjectBoundingMode)

        self._radial = QtGui.QRadialGradient(.5, .5, .5)
        self._radial.setCoordinateMode(QtGui.QGradient.ObjectBoundingMode)

        self._conical = QtGui.QConicalGradient(.5, .5, 0)
        self._conical.setCoordinateMode(QtGui.QGradient.ObjectBoundingMode)

        self._origin = origin
        distance = max(0, distance)
        self._clipRadius = min(distance, max(0, clipRadius))

        self._setColors(lightColor, darkColor)
        self._setDistance(distance)

    def setColors(self, color1, color2):
        # if (isinstance(color1, QtCore.Qt.GlobalColor) 
        #     and isinstance(color2, QtCore.Qt.GlobalColor)):
        #     color1 = QtGui.QColor(color1)
        #     color2 = QtGui.QColor(color2)

            self._setColors(color1, color2)
            self._setDistance(self._distance)

            self.update()

    def _setColors(self, color1, color2):
        self._baseStart = color1
        self._baseStop = QtGui.QColor(color1)
        self._baseStop.setAlpha(0)

        self._shadowStart = color2
        self._shadowStop = QtGui.QColor(color2)
        self._shadowStop.setAlpha(0)

        self.lightSideStops = [(0, self._baseStart), (1, self._baseStop)]
        self.shadowSideStops = [(0, self._shadowStart), (1, self._shadowStop)]
        self.cornerStops = [(0, self._shadowStart), (.25, self._shadowStop),
                            (.75, self._shadowStop), (1, self._shadowStart)]

        self._setOrigin(self._origin)

    def distance(self):
        return self._distance

    def setDistance(self, distance):
        if distance == self._distance:
            return

        self._setDistance(distance)
        self.updateBoundingRect()

    def _getCornerPixmap(self, rect, grad1, grad2=None):
        pm = QtGui.QPixmap(self._distance + self._clipRadius, 
                           self._distance + self._clipRadius)
        pm.fill(QtCore.Qt.GlobalColor.transparent)

        qp = QtGui.QPainter(pm)
        if self._clipRadius > 1:
            path = QtGui.QPainterPath()
            size = self._clipRadius * 2 - 1
            el = QtCore.QRectF(0, 0, size, size)
            path.addEllipse(rect)
            qp.setClipPath(path)
        qp.fillRect(rect, grad1)
        if grad2:
            qp.setCompositionMode(qp.CompositionMode.CompositionMode_SourceAtop)
            qp.fillRect(rect, grad2)
        qp.end()
        return pm

    def _setDistance(self, distance):
        distance = max(1, distance)
        self._distance = distance
        if self._clipRadius > distance:
            self._clipRadius = distance
        distance += self._clipRadius
        r = QtCore.QRectF(0, 0, distance * 2, distance * 2)
        lightSideStops = self.lightSideStops[:]
        shadowSideStops = self.shadowSideStops[:]
        if self._clipRadius:
            gradStart = self._clipRadius / (self._distance + self._clipRadius)
            lightSideStops[0] = (1, lightSideStops[0][1])
            lightSideStops[1] = (gradStart, lightSideStops[1][1])
            shadowSideStops[1] = (gradStart, shadowSideStops[1][1])
            shadowSideStops[0] = (1, shadowSideStops[0][1])
        else:
            lightSideStops[0] = (1, lightSideStops[0][1])
            lightSideStops[1] = (0, lightSideStops[1][1])
            shadowSideStops[1] = (0, shadowSideStops[1][1])
            shadowSideStops[0] = (1, shadowSideStops[0][1])

        # create the 4 corners as if the light source was top-left
        self._radial.setStops(lightSideStops)
        topLeft = self._getCornerPixmap(r, self._radial)

        self._conical.setAngle(359.9)
        self._conical.setStops(self.cornerStops)
        topRight = self._getCornerPixmap(r.translated(-distance, 0), 
                                         self._radial, self._conical)

        self._conical.setAngle(270)
        self._conical.setStops(self.cornerStops)
        bottomLeft = self._getCornerPixmap(r.translated(0, -distance), 
                                           self._radial, self._conical)

        self._radial.setStops(shadowSideStops)
        bottomRight = self._getCornerPixmap(r.translated(-distance, -distance), 
                                            self._radial)

        # rotate the images according to the actual light source
        images = topLeft, topRight, bottomRight, bottomLeft

        shift = self._cornerShift.index(self._origin)
        if shift:
            transform = QtGui.QTransform().rotate(shift * 90)
            for img in images:
                img.swap(img.transformed(transform, QtCore.Qt.SmoothTransformation))

        # and reorder them if required
        self.topLeft, self.topRight, self.bottomRight, self.bottomLeft = images[-shift:] + images[:-shift]

    def origin(self):
        return self._origin

    def setOrigin(self, origin):
        origin = QtCore.Qt.Corner(origin)
        if origin == self._origin:
            return

        self._setOrigin(origin)
        self._setDistance(self._distance)
        self.update()

    def _setOrigin(self, origin):
        self._origin = origin

        gradients = self._leftGradient, self._topGradient, self._rightGradient, self._bottomGradient
        stops = self.lightSideStops, self.lightSideStops, self.shadowSideStops, self.shadowSideStops

        # assign color stops to gradients based on the light source position
        shift = self._cornerShift.index(self._origin)
        for grad, stops in zip(gradients, stops[-shift:] + stops[:-shift]):
            grad.setStops(stops)

    def clipRadius(self):
        return self._clipRadius

    def setClipRadius(self, radius):
        if radius == self._clipRadius:
            return

        self._setClipRadius(radius)
        self.update()

    def _setClipRadius(self, radius):
        radius = min(self._distance, max(0, int(radius)))
        self._clipRadius = radius
        self._setDistance(self._distance)

    def boundingRectFor(self, rect):
        d = self._distance
        return rect.adjusted(-d, -d, d, d)

    def draw(self, qp):

        restoreTransform = qp.worldTransform()

        qp.setPen(QtCore.Qt.NoPen)
        x, y, width, height = self.sourceBoundingRect(QtCore.Qt.DeviceCoordinates).getRect()
        right = x + width
        bottom = y + height
        clip = self._clipRadius
        doubleClip = clip * 2

        if self._clipRadius:
            path = QtGui.QPainterPath()
            source = self.sourcePixmap(QtCore.Qt.DeviceCoordinates)
            sourceBoundingRect = self.sourceBoundingRect(QtCore.Qt.DeviceCoordinates)
            qp.save()
            qp.setTransform(QtGui.QTransform())
            path.addRoundedRect(sourceBoundingRect, self._clipRadius, self._clipRadius)
            qp.setClipPath(path)
            qp.drawPixmap(sourceBoundingRect.x() - self._distance, sourceBoundingRect.y() - self._distance, source)
            qp.restore()
        else:
            path = QtGui.QPainterPath()
            source = self.sourcePixmap(QtCore.Qt.DeviceCoordinates)
            sourceBoundingRect = self.sourceBoundingRect(QtCore.Qt.DeviceCoordinates)
            qp.save()
            qp.setTransform(QtGui.QTransform())
            path.addRect(sourceBoundingRect.x(), sourceBoundingRect.y(), sourceBoundingRect.width(),
                         sourceBoundingRect.height())
            qp.setClipPath(path)
            qp.drawPixmap(sourceBoundingRect.x() - self._distance, sourceBoundingRect.y() - self._distance, source)
            qp.restore()

        qp.setWorldTransform(QtGui.QTransform())
        leftRect = QtCore.QRectF(x, y + clip + self._distance, self._distance,
                                 height - doubleClip - self._distance * 2)
        qp.setBrush(self._leftGradient)
        qp.drawRect(leftRect)

        topRect = QtCore.QRectF(x + clip + self._distance, y, width - doubleClip - self._distance * 2,
                                self._distance)
        qp.setBrush(self._topGradient)
        qp.drawRect(topRect)

        rightRect = QtCore.QRectF(right - self._distance, y + clip + self._distance, self._distance,
                                  height - doubleClip - self._distance * 2)
        qp.setBrush(self._rightGradient)
        qp.drawRect(rightRect)

        bottomRect = QtCore.QRectF(x + clip + self._distance, bottom - self._distance,
                                   width - doubleClip - self._distance * 2, self._distance)
        qp.setBrush(self._bottomGradient)
        qp.drawRect(bottomRect)

        qp.drawPixmap(x, y, self.topLeft)
        qp.drawPixmap(right - clip - self._distance, y, self.topRight)
        qp.drawPixmap(right - clip - self._distance, bottom - clip - self._distance, self.bottomRight)
        qp.drawPixmap(x, bottom - clip - self._distance, self.bottomLeft)

        qp.setWorldTransform(restoreTransform)



from ..core import QObjectBase

class OutsideNeumorphismEffect(QObjectBase, _OutsideNeumorphismEffect):
    def __init__(self,
                 *,
                 light_color: QtGui.QColor = QtGui.QColor("#FFFFFF"),
                 dark_color: QtGui.QColor = QtGui.QColor("#7d7d7d"),
                 clip_radius: int = 4,
                 distance: int = 4,
                 origin: QtCore.Qt.Corner = QtCore.Qt.Corner.TopLeftCorner,
                 **kwargs) -> QtWidgets.QWidget:
        _OutsideNeumorphismEffect.__init__(self)
        super().__init__(**kwargs)

        if clip_radius is not None:
            self.setClipRadius(clip_radius)

        if light_color is not None and dark_color is not None:
            self.setColors(light_color,
                           dark_color)

        if distance is not None:
            self.setDistance(distance)

        if origin is not None:
            self.setOrigin(origin)


class InsideNeumorphismEffect(QObjectBase, _InsideNeumorphismEffect):
    def __init__(self,
                 *,
                 light_color: QtGui.QColor = QtGui.QColor("#FFFFFF"),
                 dark_color: QtGui.QColor = QtGui.QColor("#7d7d7d"),
                 clip_radius: int = 4,
                 distance: int = 4,
                 origin: QtCore.Qt.Corner = QtCore.Qt.Corner.BottomRightCorner,
                 **kwargs):
        _InsideNeumorphismEffect.__init__(self)
        super().__init__(**kwargs)

        if clip_radius is not None:
            self.setClipRadius(clip_radius)

        if light_color is not None and dark_color is not None:
            self.setColors(light_color,
                           dark_color)

        if distance is not None:
            self.setDistance(distance)

        if origin is not None:
            self.setOrigin(origin)


#!/usr/bin/python
# -*- coding: utf-8 -*-

from enum import IntEnum

from PySide6.QtCore import (
    QAbstractAnimation, 
    QCoreApplication, 
    QEvent,
    QTime
)

ScrollBarFadeOutDuration = 200.0
ScrollBarFadeOutDelay = 450.0
StyleAnimationUpdate = 213


class StyleAnimation(QAbstractAnimation):

    FrameRate = IntEnum(
        'FrameRate',
        ['DefaultFps', 
         'SixtyFps', 
         'ThirtyFps', 
         'TwentyFps', 
         'FifteenFps'])

    def __init__(self, *args, **kwargs):
        super(StyleAnimation, self).__init__(*args, **kwargs)
        self._delay = 0
        self._duration = -1
        self._startTime = QTime.currentTime()
        self._fps = self.FrameRate.ThirtyFps
        self._skip = 0

    def target(self):
        return self.parent()

    def duration(self):
        return self._duration

    def setDuration(self, duration):
        self._duration = duration

    def delay(self):
        return self._delay

    def setDelay(self, delay):
        self._delay = delay

    def startTime(self):
        return self._startTime

    def setStartTime(self, time):
        self._startTime = time

    def frameRate(self):
        return self._fps

    def setFrameRate(self, fps):
        self._fps = fps

    def updateTarget(self):
        event = QEvent(QEvent.Type(StyleAnimationUpdate))
        event.setAccepted(False)
        QCoreApplication.sendEvent(self.target(), event)
        if not event.isAccepted():
            self.stop()

    def start(self):
        self._skip = 0
        super(StyleAnimation, self).start(QAbstractAnimation.DeletionPolicy.DeleteWhenStopped)

    def isUpdateNeeded(self):
        return self.currentTime() > self._delay

    def updateCurrentTime(self, _):
        self._skip += 1
        if self._skip >= self._fps:
            self._skip = 0
            if self.parent() and self.isUpdateNeeded():
                self.updateTarget()


class ProgressStyleAnimation(StyleAnimation):
    def __init__(self, speed, *args, **kwargs):
        super(ProgressStyleAnimation, self).__init__(*args, **kwargs)
        self._speed = speed
        self._step = -1

    def animationStep(self):
        return self.currentTime() / (1000.0 / self._speed)

    def progressStep(self, width):
        step = self.animationStep()
        progress = (step * width / self._speed) % width
        if (((step * width / self._speed) % (2 * width)) >= width):
            progress = width - progress
        return progress

    def speed(self):
        return self._speed

    def setSpeed(self, speed):
        self._speed = speed

    def isUpdateNeeded(self):
        if super(ProgressStyleAnimation, self).isUpdateNeeded():
            current = self.animationStep()
            if self._step == -1 or self._step != current:
                self._step = current
                return True
        return False



from .overload import singledispatchmethod
from .exception_handler import exceptionHandler
from .auto_wrap import TextWrap
from ._ensure_thread import CallCallable, ensure_object_thread, ensure_main_thread
from ._message_handler import MessageHandler

import uuid

def UniqueIdentifier() -> str:
    return uuid.uuid4().hex


#!/usr/bin/python3

from PySide6.QtCore import QFile
from string import Template
from typing import Union
import json
import os


class _QssTemplate(Template):
    delimiter = "--"


class QStyleSheet:
    '''
    Label(
        qss=QStyleSheet(
            qss={
                "background-color": "--themeColor",
                "color": "white",
                "font-size": "24px"
            },
            vars={
                "themeColor": "blue"
            }
        )
    )
    
    :param: qss: 
        - dict: {"background-color": "white"}
        - str: "QWidget { background-color: white;}"
        - str: ":/path/to/resoucre/file.ext", 
                and "/path/to/local/file.ext",
                and "/path/to/local/file.json"

    :Qss result
        QLabel {
            background-color: blue;
            color: white;
            font-size: 24px;
        }
    '''

    def __init__(self, 
                 qss: Union[str, dict],
                 vars: dict = None):

        self._qss = qss
        self._vars = vars
        self._temp = _QssTemplate("")
        self._parent = None

    def _set_parent(self, parent: object):
        self._parent = parent

    def _get_qss_from_str(self, qss: str) -> str:
        if qss.startswith(":") or os.path.exists(qss):
            file = QFile(qss)
            file.open(QFile.OpenModeFlag.ReadOnly)
            data = str(file.readAll(), "utf-8")

            # add .json support
            if qss.lower().endswith(".json"):
                data = self._get_qss_from_dict(json.loads(data))

            file.close()
            return data
        return qss

    def _get_qss_parent(self) -> str:
        _parent = self._parent.objectName().strip()
        if not _parent:
            _parent = self._parent.__class__.__base__.__name__
            return _parent
        return "#" + _parent

    def _get_qss_from_dict(self, qss: dict) -> str:
        style_sheet = ""
        _obj_style = ""

        for k, v in qss.items():
            if isinstance(v, dict) and not k.startswith(":"):
                raise ValueError("Invalid Qss parent!")

            if k.startswith(("#", ".", "*")):
                raise ValueError("Invalid Qss key value!")

            if isinstance(v, dict):
                style_sheet += "%s%s {\n" % (self._get_qss_parent(), k)
                for sk, sv in v.items():
                    style_sheet += f"\t{sk}: {sv};\n"
                style_sheet += "}\n"
            else:
                _obj_style += f"\t{k}: {v};\n"

        style_sheet += "%s {\n" % self._get_qss_parent()
        style_sheet += _obj_style
        style_sheet += "}\n"

        return style_sheet

    def _set_qss(self, qss: Union[str, dict]) -> None:
        if qss:
            _style = (self._get_qss_from_str(qss)
                    if isinstance(qss, str) 
                    else self._get_qss_from_dict(qss))

            if self._vars:
                self._temp.template = _style
                self._temp.safe_substitute(_style)
                _style = self._temp.safe_substitute(self._vars)

            self._parent.setStyleSheet(_style)

    def update_vars(self, vars: dict) -> None:
        self._vars = vars
        self._set_qss(self._qss)

    def update_qss(self,
                   qss: Union[str, dict],
                   *,
                   save: bool = False) -> None:

        if self._qss is not None:
            if (isinstance(qss, dict) 
                and isinstance(self._qss, dict)):
                _qss = self._qss.copy()
                _qss.update(qss)
                self._qss = qss if save else self._qss
                _style = self._get_qss_from_dict(_qss)
            else:
                self._qss = qss if save else self._qss
                _style = self._get_qss_from_str(qss)
        else:
            _style = (
                self._get_qss_from_dict(qss)
                if isinstance(qss, dict)
                else
                self._get_qss_from_str(qss)
            )

            if save:
                self._qss = qss

        if self._vars:
            self._temp.template = _style
            self._temp.safe_substitute(_style)
            _style = self._temp.safe_substitute(self._vars)

        self._parent.setStyleSheet(_style)

    def restore_qss(self) -> None:
        if self._qss is not None:
            if isinstance(self._qss, dict):
                _style = self._get_qss_from_dict(self._qss)
            else:
                _style = self._get_qss_from_str(self._qss)

            if self._vars:
                self._temp.template = _style
                self._temp.safe_substitute(_style)
                _style = self._temp.safe_substitute(self._vars)

            self._parent.setStyleSheet(_style)
        else:
            self._parent.setStyleSheet("")

    def update(self, 
               qss: Union[str, dict], 
               vars: dict = None, 
               *,
               save: bool = False) -> None:

        if vars is not None:
            self.update_vars(vars)

        self.update_qss(qss, save=save)

    def restore(self) -> None:
        self.restore_qss()


import os

from enum import Enum
from typing import Any, Union
from ..enums.env_vars import EnvVars


class EnvVar:
    """
    e.g, EnvVar.set(EnvVars.scale_factor, "0")
    e.g, EnvVar.set("QT_SCALE_FACTOR", "0")
    """

    @staticmethod
    def get(key: str, default: Any = None) -> Any:
        return os.environ.get(key, default)

    @staticmethod
    def set(key: Union[EnvVars, str], value: Any) -> None:
        os.environ[key.value 
                   if isinstance(key, EnvVars) 
                   else key] = (value.value 
                                if isinstance(value, Enum) 
                                else value)


from PySide6.QtWidgets import QFrame, QScrollArea, QWidget
from PySide6.QtCore import Qt
from ..core import WidgetBase


class ScrollArea(WidgetBase, QScrollArea):
    def __init__(self, 
                 *,
                 child: QWidget = None,
                 **kwargs):
        QScrollArea.__init__(self)
        super().__init__(**kwargs)

        self.setWidgetResizable(True)
        self.setUpdatesEnabled(True)

        self.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.setSizeAdjustPolicy(QScrollArea.SizeAdjustPolicy.AdjustToContents)

        self.setFrameShadow(QFrame.Shadow.Plain)
        self.setFrameShape(QFrame.Shape.NoFrame)

        if child is not None:
            self.setWidget(child)


from .ui_loader import UILoader
from .qt_file_open import File
from .elided_text import ElidedText
from .copy_progress import CopyProgress
from .smooth_scroll import SmoothScroll
from .icon import Icon
from .env_var import EnvVar
from .color import Color
from .picker import ColorPicker
from .modifiers import Modifiers
from .size_policy import SizePolicy
from .alignment import Alignment
from .action import Action
from .system_tray import SystemTray

from .painters import (
    StatusEdgePaint
)

from ._widgets import (
    MenuSection,
    MenuSeparator,
    MenuSimpleAction
)


#!/usr/bin/python3

class WindowsInit:
    def __init__(self) -> None:
        ...

    def set_app_id(self, 
                   company: str, 
                   product: str, 
                   subproduct: str, 
                   version: str) -> str:
        try:
            from ctypes import windll  # Only exists on Windows.
            windll.shell32.SetCurrentProcessExplicitAppUserModelID(f"{company}.{product}.{subproduct}.{version}")
        except ImportError:
            pass


from PySide6.QtWidgets import QWidget, QLayout
from ..core import WidgetBase


class Widget(WidgetBase, QWidget):
    def __init__(self,
                 *,
                 child: QWidget | QLayout = None,
                 **kwargs):
        QWidget.__init__(self)
        super().__init__(**kwargs)

        self._set_child(child)

    def _set_child(self, child: QWidget | QLayout):
        if child is not None:
            if isinstance(child, QWidget):
                child.setParent(self)

            elif isinstance(child, QLayout):
                child.setProperty("parent", self)
                self.setLayout(child)
            else:
                raise ValueError("the 'child' argument must be one of the QWidget or QLayout instance.")


from typing import Union
from PySide6.QtGui import QShowEvent, QCloseEvent
from PySide6.QtCore import Signal
from PySide6.QtWidgets import (
    QMainWindow,
    QSystemTrayIcon, 
    QWidget, 
    QToolBar,
    QStatusBar,
    QDockWidget,
    QLayout
)
from ...core import WidgetBase


class MainWindow(WidgetBase, QMainWindow):
    startup_changed = Signal()

    def __init__(self, 
                 *,
                 home: Union[QWidget, QLayout] = None,
                 tool_bar: QToolBar = None,
                 status_bar: QStatusBar = None,
                 dock_widget: QDockWidget = None,
                 sys_tray: QSystemTrayIcon = None,
                 **kwargs):
        QMainWindow.__init__(self)
        super().__init__(**kwargs)

        self.__is_startup = False

        self.setAnimated(True)
        self.setUpdatesEnabled(True)

        if home is not None:
            self._set_home(home)
        
        if tool_bar is not None:
            self._set_tool_bar(tool_bar)

        if status_bar is not None:
            self._set_status_bar(status_bar)

        if dock_widget is not None:
            self._set_dock_widget(dock_widget)

        if sys_tray is not None:
            sys_tray.setParent(self)
            sys_tray.show()

    def closeEvent(self, event: QCloseEvent) -> None:
        super().closeEvent(event)
        self.__is_startup = False

    def showEvent(self, event: QShowEvent) -> None:
        super().showEvent(event)
        if not self.__is_startup:
            self.__is_startup = True
            self.startup_changed.emit()

    def _set_home(self, home: QWidget) -> None:
        if isinstance(home, QLayout):
            self.centralwidget = QWidget(self)
            self.centralwidget.setObjectName("centralwidget")
            home.setProperty("parent", self.centralwidget)
            self.centralwidget.setLayout(home)
            self.setCentralWidget(self.centralwidget)

        elif isinstance(home, QWidget):
            home.setParent(self)
            self.setCentralWidget(home)

        else:
            raise ValueError("the 'home' argument must be one of \
                the QWidget or QLayout instance.")

    def _set_tool_bar(self, tool_bar: QToolBar) -> None:
        self.addToolBar(tool_bar)

    def _set_status_bar(self, status_bar: QStatusBar) -> None:
        self.setStatusBar(status_bar)

    def _set_dock_widget(self, dock_widget: QDockWidget) -> None:
        self.addDockWidget(dock_widget)



#!/usr/bin/python3

import os
import sys

from typing import Sequence, Union, Iterable
from PySide6.QtGui import QFontDatabase
from PySide6.QtCore import QResource, Qt, Signal, qRegisterResourceData
from PySide6.QtWidgets import QApplication, QStyleFactory, QWidget

from ..core import QObjectBase
from ..utils._import import Imports
from ..utils.overload import staticproperty
from .._rc.resource import qInitResources


class Application(QObjectBase, QApplication):
    on_inactive = Signal()
    on_active = Signal()
    on_hidden = Signal()
    on_suspend = Signal()

    def __init__(self,
                 *,
                 arg: Sequence[str] = None,
                 resources: list[Union[str, tuple]] = None,
                 fonts: list[str] = None,
                 **kwargs):
        QApplication.__init__(self, arg or [])
        super().__init__(**kwargs)

        # init Qtica default resource
        qInitResources()

        if resources is not None:
            self._set_resources(resources)

        if fonts is not None:
            self._set_fonts(fonts)

        ## it's default enabled by the developer
        self.setAttribute(Qt.ApplicationAttribute.AA_UseHighDpiPixmaps, True)
        self.setAttribute(Qt.ApplicationAttribute.AA_EnableHighDpiScaling, True)

        self.applicationStateChanged.connect(self._applicationStateChanged)
    
    def _applicationStateChanged(self, event):
        if event == Qt.ApplicationState.ApplicationInactive:
            self.on_inactive.emit()
        elif event == Qt.ApplicationState.ApplicationActive:
            self.on_active.emit()
        elif event == Qt.ApplicationState.ApplicationHidden:
            self.on_hidden.emit()
        elif event == Qt.ApplicationState.ApplicationSuspended:
            self.on_suspend.emit()

    def run(self) -> int:
        return sys.exit(self.exec())

    def list_styles(self) -> list[str]:
        return QStyleFactory.keys()

    def current_style(self) -> str:
        return self.style().name()

    def set_style(self, name: str):
        if name.lower().strip() not in [n.lower() for n in QStyleFactory.keys()]:
            raise ValueError(f"style '{name}' not found please call list_styles to \
                see all available styles in your system.")

        self.setStyle(QStyleFactory.create(name))

    def add_rcc_file(self,
                     rcc_file: str,
                     rcc_root: str = None) -> bool:

        if rcc_root is not None:
            return QResource.registerResource(rcc_file, 
                                              rcc_root)
        return QResource.registerResource(rcc_file)

    def add_font(self, fontpath: str) -> int:
        return QFontDatabase.addApplicationFont(fontpath)

    def _set_fonts(self, fonts: list[str]):
        for font in fonts:
            if not os.path.exists(font):
                raise FileNotFoundError(f"'{font}' font file not found!")

            self.add_font(fonts)

    def _set_resources(self, resources: list[Union[str, tuple]]):
        '''
        * resource.py file must contain the qInitResources method
        * Please comment qInitResources calling in the tail of resource.py the file.
        :param: resource
            str = "/path/to/file.py
            tuple = (qt_resource_struct, qt_resource_name, qt_resource_data)
        '''
        for res in resources:
            if isinstance(res, Iterable):
                qRegisterResourceData(0x03, *res)

            elif res.endswith((".py", ".pyc")):
                if not os.path.exists(res):
                    raise FileNotFoundError(f"'{res}' resource file not found!")
                try:
                    Imports.method(res, "qInitResources")()
                except Exception as err:
                    raise ValueError(f"Resource error: '{res}', {err}")

    @staticproperty
    def active_window() -> QWidget:
        return QApplication.activeWindow()

    @staticproperty
    def active_modal() -> QWidget:
        return QApplication.activeModalWidget()

    @staticproperty
    def active_popup() -> QWidget:
        return QApplication.activePopupWidget()


class App(Application):
    pass


from typing import List, Optional
from PySide6.QtGui import QColor, QResizeEvent
from PySide6.QtWidgets import QFrame, QSizePolicy, QSpacerItem, QVBoxLayout, QWidget
from .side_bar_button import SideBarButton


_BG_FRAME_STYLESHEET = """\
QFrame#_side_bar_bg_frame {{
    background-color: {bg_color};
    border: none;
    border-radius: 8px;
}}
"""


class SideBar(QWidget):
    def __init__(self, 
                 parent: Optional[QWidget] = None,
                 bg_color: QColor = QColor("#191E23")) -> None:
        super().__init__(parent)

        # Define attributes
        self._first_button = True
        self._top_buttons: List[SideBarButton] = []
        self._bottom_buttons: List[SideBarButton] = []

        # Layout for this widget
        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)

        # Background frame and its layout
        self._bg_frame = QFrame(self)
        self._bg_frame.setObjectName("_side_bar_bg_frame")
        self._bg_layout = QVBoxLayout(self._bg_frame)
        self._bg_layout.setContentsMargins(0, 15, 0, 15)
        self._bg_layout.setSpacing(0)
        self._layout.addWidget(self._bg_frame)

        # Top side bar buttons frame and its layout
        self._top_frame = QFrame(self._bg_frame)
        self._top_frame_layout = QVBoxLayout(self._top_frame)
        self._top_frame_layout.setContentsMargins(0, 0, 0, 0)
        self._top_frame_layout.setSpacing(0)
        self._bg_layout.addWidget(self._top_frame)

        # Vertical spacer between top and bottom bar
        self._vertical_spacer = QSpacerItem(0, 0, 
                                            QSizePolicy.Policy.Minimum, 
                                            QSizePolicy.Policy.Expanding)
        self._bg_layout.addItem(self._vertical_spacer)

        # Bottom side bar buttons frame and its layout
        self._bottom_frame = QFrame(self._bg_frame)
        self._bottom_frame_layout = QVBoxLayout(self._bottom_frame)
        self._bottom_frame_layout.setContentsMargins(0, 0, 0, 0)
        self._bottom_frame_layout.setSpacing(0)
        self._bg_layout.addWidget(self._bottom_frame)

        self.setBackgroundColor(bg_color)

    def setBackgroundColor(self, color: QColor("#191E23")) -> None:
        self._bg_frame.setStyleSheet(_BG_FRAME_STYLESHEET.format(bg_color=color.name()))

    def resizeEvent(self, event: QResizeEvent) -> None:
        super().resizeEvent(event)

        buttons: List[SideBarButton] = self.findChildren(SideBarButton)
        for button in buttons:
            button.setFixedSize(self.width(), 
                                self.width())

    def addButton(self, 
                  button: SideBarButton, 
                  alignment: SideBarButton.Alignment = SideBarButton.Alignment.top) -> None:

        button.setParent(self)
        button.clicked.connect(self._buttonCallback)  # type: ignore[attr-defined]

        if alignment == SideBarButton.Alignment.top:
            self._top_buttons.append(button)
            self._top_frame_layout.addWidget(button)
        elif alignment == SideBarButton.Alignment.bottom:
            self._bottom_buttons.append(button)
            self._bottom_frame_layout.addWidget(button)
        else:
            raise ValueError(
                "Invalid value for 'alignment'!"
                "Supported values are 'SideBarButton.Alignment.top' and 'SideBarButton.Alignment.bottom'."
            )

        # Select the first button as default
        if self._first_button:
            button.click()
            self._first_button = False

    def _buttonCallback(self):
        buttons: List[SideBarButton] = self.findChildren(SideBarButton)
        for button in buttons:
            if button == self.sender():
                button.setChecked(True)
            else:
                button.setChecked(False)



from typing import Any, Sequence, Tuple
from PySide6.QtCore import (
    QAbstractAnimation, 
    QByteArray, 
    QEasingCurve,
    QPoint,
    QPointF, 
    QRectF
)
from PySide6.QtGui import QPainterPath
from PySide6.QtWidgets import QWidget
from ..enums.animation import AnimationPath
from .property_animation import PropertyAnimation


class Animation(PropertyAnimation):
    def __init__(self, 
                 *,
                 child: QWidget,
                 property_name: QByteArray | bytes,
                 duration: int = None,
                 start_value: Any = None,
                 end_value: Any = None,
                 easing_curve: QEasingCurve | QEasingCurve.Type = None,
                 loop_count: int = None,
                 key_value: Tuple[float, Any] | Sequence[Tuple[float, Any]] = None,
                 path_type: AnimationPath = AnimationPath.linear,
                 direction: QAbstractAnimation.Direction = None,
                 running: bool = False,
                 **kwargs) -> QWidget:

        if path_type is not None:
            self._pathType = path_type
            self.setPath(path_type)

        return super().__init__(child, property_name, duration, 
                                start_value, end_value, easing_curve, 
                                loop_count, key_value, direction, 
                                running, **kwargs)

    def setPath(self, pathType):
        self._pathType = pathType
        self._path = QPainterPath()

    def updateCurrentTime(self, currentTime):
        if self._pathType == AnimationPath.circle:
            if self._path.isEmpty():                
                end = self.endValue()
                start = self.startValue()

                assert (isinstance(end, (QPoint, QPointF, QPainterPath.ElementType))
                        or isinstance(start, (QPoint, QPointF, QPainterPath.ElementType))), \
                            "startValue, endValue should be " +\
                                "QPoint | QRectF | QPainterPath.ElementType" 

                self._path.moveTo(start)
                self._path.addEllipse(QRectF(start, end))

            dura = self.duration()
            if dura == 0:
                progress = 1.0
            else:
                progress = (((currentTime - 1) % dura) + 1) / float(dura)

            eased_progress = self.easingCurve().valueForProgress(progress)
            if eased_progress > 1.0:
                eased_progress -= 1.0
            elif eased_progress < 0:
                eased_progress += 1.0

            pt = self._path.pointAtPercent(eased_progress)

            self.updateCurrentValue(pt)
            self.valueChanged.emit(pt)
        else:
            super(Animation, self).updateCurrentTime(currentTime)


#-----------------------------------------------------------------------------
#  Copyright (C) 2019 Alberto Sottile
#
#  Distributed under the terms of the 3-clause BSD License.
#-----------------------------------------------------------------------------

import typing
from ...enums.theme import Theme


def theme():
    return None
        
def isDark():
    return None
    
def isLight():
    return None

def listener(callback: typing.Callable[[Theme], None]) -> None:
    raise NotImplementedError()



from PySide6.QtCore import Qt, QEvent
from PySide6.QtGui import QPainter, QColor
from PySide6.QtWidgets import QGridLayout, QWidget
from .tails import TeachingTipManager
from ...enums import TeachingTipTailPositions
from ...utils.theme_detect import isDark
from ...core import WidgetBase


class _TeachingTip(QWidget):
    def __init__(self, 
                 view: QWidget,
                 target: QWidget,
                 tail_position: TeachingTipTailPositions = TeachingTipTailPositions.bottom,
                 delete_on_close: bool = False,
                 radius: int = 6,
                 parent: QWidget = None
                ) -> None:
        super().__init__(parent=parent)

        self.view = view
        self.target = target
        self.delete_on_close = delete_on_close
        self.manager = TeachingTipManager.make(tail_position)
        self.manager.radius = radius

        self._layout = QGridLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self.manager.doLayout(self)

        self._layout.addWidget(view)

        # set style
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        self.setAttribute(Qt.WidgetAttribute.WA_StyledBackground)
        self.setWindowFlags(Qt.WindowType.Tool | Qt.WindowType.FramelessWindowHint)

        if parent and parent.window():
            parent.window().installEventFilter(self)

        if self.target is not None:
            self.target.resizeEvent = lambda e: self.update_pos()
            self.target.moveEvent = lambda e: self.update_pos()
            self.target.window().moveEvent = lambda e: self.update_pos()

    def tooltip_pos(self):
        return self.manager.position(self)

    def set_target(self, target: QWidget):
        self.target = target
        self.target.resizeEvent = lambda e: self.update_pos()
        self.target.moveEvent = lambda e: self.update_pos()
        self.target.window().moveEvent = lambda e: self.update_pos()

    def update_pos(self):
        self.move(self.manager.position(self))

    def showEvent(self, e):
        self.update_pos()
        super().showEvent(e)

    def closeEvent(self, e):
        if self.delete_on_close:
            self.deleteLater()
        super().closeEvent(e)

    def eventFilter(self, obj, e: QEvent):
        if self.parent() and obj is self.parent().window():
            if e.type() in [QEvent.Type.Resize, 
                            QEvent.Type.WindowStateChange, 
                            QEvent.Type.Move]:
                self.move(self.manager.position(self))
        return super().eventFilter(obj, e)

    def paintEvent(self, e, scale: float = None):
        super().paintEvent(e)

        painter = QPainter(self)
        painter.setRenderHints(QPainter.RenderHint.Antialiasing)

        painter.setBrush(
            QColor(40, 40, 40) 
            if isDark() 
            else QColor(248, 248, 248))
        
        painter.setPen(
            QColor(23, 23, 23) 
            if isDark() 
            else QColor(195, 195, 195))

        self.manager.draw(self, painter)


class TeachingTip(WidgetBase, _TeachingTip):
    def __init__(self, 
                 *,
                 child: QWidget,
                 target: QWidget,
                 tail_position: TeachingTipTailPositions = TeachingTipTailPositions.bottom,
                 delete_on_close: bool = False,
                 radius: int = 6,
                 **kwargs):
        _TeachingTip.__init__(self, 
                              child, 
                              target, 
                              tail_position, 
                              delete_on_close,
                              radius)
        super().__init__(**kwargs)


from PySide6.QtWidgets import QSizePolicy, QWidget
from ..core import ToolBase


class SizePolicy(ToolBase, QSizePolicy):
    def __init__(self,
                 *,
                 child: QWidget,
                 horizontal: QSizePolicy.Policy = None,
                 vertical: QSizePolicy.Policy = None,
                 stretch: tuple[int, int] = None,
                 type: QSizePolicy.ControlType = None,
                 **kwargs) -> QWidget:
        QSizePolicy.__init__(self)
        super().__init__(**kwargs)

        if type is not None:
            self.setControlType(type)

        if horizontal is not None:
            self.setHorizontalPolicy(horizontal)

        if vertical is not None:
            self.setVerticalPolicy(vertical)

        self.setHorizontalStretch(stretch[0] if stretch is not None else 0)
        self.setVerticalStretch(stretch[-1] if stretch is not None else 0)

        self.setHeightForWidth(child.sizePolicy().hasHeightForWidth())
        self.setWidthForHeight(child.sizePolicy().hasWidthForHeight())
        # sizePolicy.setRetainSizeWhenHidden(True)

        child.setSizePolicy(self)

        return child


from typing import Any, Union, Optional
from PySide6.QtCore import Qt
from PySide6.QtGui import QColor, QRgba64
from ..core.base import BehaviorDeclarative
from ..enums.colors import Colors


class Color(BehaviorDeclarative):
    '''
    e.g Color(Colors.red) -> QColor
    '''
    def __init__(self,
                 color: Union[QColor, 
                              QRgba64, 
                              Any,
                              Qt.GlobalColor, 
                              Colors,
                              str, 
                              int, 
                              tuple[int, int, int, Optional[int]],
                              tuple[QColor.Spec, 
                                    int, 
                                    int, 
                                    int, 
                                    int, 
                                    Optional[int]]]) -> QColor:

        if isinstance(color, Colors):
            return color.value

        if isinstance(color, (list, tuple)):
            return QColor(*color)

        if type(color) == QColor:
            return color

        return QColor(color)


from typing import Union
from PySide6.QtGui import QAction, QIcon, QPixmap
from PySide6.QtWidgets import QLineEdit
from ..core import WidgetBase


class LineEdit(WidgetBase, QLineEdit):
    def __init__(self,
                 *,
                 password_mode: bool = False,
                 password_show_icon: Union[QIcon, QPixmap] = None,
                 password_hide_icon: Union[QIcon, QPixmap] = None,
                 **kwargs):
        QLineEdit.__init__(self)
        super().__init__(**kwargs)

        self._password_show_icon = password_show_icon
        self._password_hide_icon = password_hide_icon

        if password_mode:
            self._set_password_mode()

    def _toggle_password_echo_mode_icon(self):
        if self.echoMode() == QLineEdit.EchoMode.Password:
            self.setEchoMode(QLineEdit.EchoMode.Normal)
            # set eye close
            self._password_echo_mode_action.setIcon(self._password_hide_icon)
        else:
            self.setEchoMode(QLineEdit.EchoMode.Password)
            # set eye open
            self._password_echo_mode_action.setIcon(self._password_show_icon)

    def _set_password_mode(self):
        self.setEchoMode(QLineEdit.EchoMode.Password)
        self._password_echo_mode_action = QAction(icon=self._password_show_icon)
        self._password_echo_mode_action.triggered.connect(self._toggle_password_echo_mode_icon)
        self.addAction(self._password_echo_mode_action, 
                       QLineEdit.ActionPosition.TrailingPosition)

    def _set_normal_mode(self):
        self.setEchoMode(QLineEdit.EchoMode.Normal)
        self._password_echo_mode_action.deleteLater()

    def toggle_password_mode(self, enable: bool = False):
        if enable:
            return self._set_password_mode()
        return self._set_normal_mode


#!/usr/bin/pythone3

from enum import IntEnum
from PySide6.QtWidgets import QApplication, QVBoxLayout, QWidget
from PySide6.QtGui import QColor, QIcon, QMouseEvent, QPaintEvent, QPainter
from PySide6.QtCore import QRect, QRectF, Qt
from dataclasses import dataclass
from ..core import WidgetBase
from ..tools import Icon


class ThemeSwitchButton(WidgetBase, QWidget):
    '''
    :param: current_mode: False = Night, True = Light
    '''

    @dataclass
    class NightMode:
        bg_color: QColor = QColor("#303030")
        indicator_color: QColor = QColor("#FEFFFE")
        icon: QIcon = None

        @property
        def mode(self) -> bool:
            return False

    @dataclass
    class LightMode:
        bg_color: QColor = QColor("#FF9400")
        indicator_color: QColor = QColor("#FEFFFE")
        icon: QIcon = None

        @property
        def mode(self) -> bool:
            return True

    class Style(IntEnum):
        outline: int = 0
        inline: int = 1

    def __init__(self,
                 *,
                 night_mode: NightMode = None,
                 light_mode: LightMode = None,
                 style: Style = Style.inline,
                 current_mode: bool = False,
                 radius: int = 12,
                 **kwargs):
        QWidget.__init__(self)
        super().__init__(**kwargs)

        color = (Qt.GlobalColor.white
                 if style == ThemeSwitchButton.Style.inline 
                 else Qt.GlobalColor.black)

        self._night_mode = (night_mode 
                            if night_mode is not None 
                            else ThemeSwitchButton.NightMode(
                                icon=Icon(":/Qtica/icons/night.svg",
                                          color=color)))

        self._light_mode = (light_mode 
                            if light_mode is not None 
                            else ThemeSwitchButton.LightMode(
                                icon=Icon(":/Qtica/icons/light.svg",
                                          color=color)))

        self._current_mode = (self._light_mode
                              if current_mode
                              else self._night_mode)
        self._style = style
        self._radius = radius

        self.setContentsMargins(0, 0, 0, 0)
        self.setMinimumHeight(25)
        self.setMaximumHeight(40)

    @property
    def _is_outline(self) -> bool:
        return self._style == ThemeSwitchButton.Style.outline

    @property
    def _is_inline(self) -> bool:
        return self._style == ThemeSwitchButton.Style.inline

    @property
    def avg_size(self) -> int:
        return self.height() / 2

    def _get_rect(self, state: bool):
        _left = (
            self.width() - (self.avg_size + (self.avg_size / 2))
            if state
            else self.avg_size / 2
        )

        return QRect(_left,
                     self.height() / 2 - (self.avg_size / 2),
                     self.avg_size,
                     self.avg_size)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        self._current_mode = (self._night_mode
                              if self._current_mode.mode
                              else self._light_mode)
        self.update()
        return super().mousePressEvent(event)

    def paintEvent(self, event: QPaintEvent) -> None:
        qp = QPainter(self)
        qp.setRenderHint(QPainter.RenderHint.Antialiasing)
        bg_rect = QRect(0,
                        0,
                        self.width() -  1,
                        self.height() - 1)

        if self._is_outline:
            icon_rect = self._get_rect(self._current_mode.mode)
            indicator_rect = QRectF(icon_rect.left() - (7 / 2),
                                   icon_rect.top() - (7 / 2),
                                   icon_rect.width() + 7,
                                   icon_rect.height() + 7)
        else:
            icon_rect = self._get_rect(not self._current_mode.mode)
            indicator_rect = self._get_rect(self._current_mode.mode)
            indicator_rect = QRectF(indicator_rect.left() - 2.5,
                                   indicator_rect.top() - 2.5,
                                   indicator_rect.width() + 5,
                                   indicator_rect.height() + 5)

        ## draw background
        qp.setPen(QColor(Qt.GlobalColor.transparent))
        qp.setBrush(self._current_mode.bg_color)
        qp.drawRoundedRect(bg_rect, 
                           self._radius, 
                           self._radius)

        ## draw indicator
        # draw color
        qp.setPen(Qt.GlobalColor.transparent)
        qp.setBrush(self._current_mode.indicator_color)
        qp.drawEllipse(indicator_rect)

        # draw icon
        self._current_mode.icon.paint(qp, 
                                      icon_rect,
                                      Qt.AlignmentFlag.AlignCenter)

        if self._style == ThemeSwitchButton.Style.outline:
            icon = (
                self._night_mode.icon
                if self._current_mode.mode
                else self._light_mode.icon
            )
            icon.paint(qp,
                       self._get_rect(not self._current_mode.mode),
                       Qt.AlignmentFlag.AlignCenter)



class __Test(QWidget):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)

        self._layout = QVBoxLayout(self)
        self.switch_btn = ThemeSwitchButton(style=ThemeSwitchButton.Style.outline)
        self._layout.addWidget(self.switch_btn)

def main():
    import sys
    app = QApplication(sys.argv)
    window = __Test()
    window.resize(100, 50)
    window.show()
    sys.exit(app.exec())

if __name__ == "__main__":
    main()


#!/usr/bin/python3


from typing import Union
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QMenu
from ..tools._widgets.menu import MenuSimpleAction, MenuSection, MenuSeparator
from ..core import WidgetBase


class Menu(WidgetBase, QMenu):
    def __init__(self, 
                 *,
                 children: Union[str, int],
                 **kwargs):
        QMenu.__init__(self)
        super().__init__(**kwargs)

        self._set_children(children)

    def _set_children(self, children):
        if children is not None:
            for child in children:
                if isinstance(child, MenuSeparator):
                    self.addSeparator()

                elif isinstance(child, MenuSection):
                    self.addSection(*child)

                elif isinstance(child, QMenu):
                    child.setParent(self)
                    self.addMenu(child)

                elif isinstance(child, (QAction, MenuSimpleAction)):
                    child.setParent(self)
                    self.addAction(child)


from PySide6.QtCore import QObject
from PySide6.QtUiTools import QUiLoader
from PySide6.QtWidgets import QWidget
from .qt_file_open import File
from ..core.base import BehaviorDeclarative
import os


class UILoader(BehaviorDeclarative):
    def __init__(self,
                 file: str, 
                 parent: QObject = None) -> QWidget:

        loader = QUiLoader(parent)
        if not os.path.exists(file):
            return loader.load(file, parent)

        with File(file, File.OpenModeFlag.ReadOnly) as fr:
            return loader.load(fr, parent)


#!/usr/bin/python3

from PySide6.QtGui import QAction
from ..core import QObjectBase


class Action(QObjectBase, QAction):
    def __init__(self, **kwargs):
        QAction.__init__(self)
        super().__init__(**kwargs)


from typing import Union
from PySide6.QtGui import QIcon, QIconEngine, QImage, QPixmap, QPainter, QColor
from PySide6.QtWidgets import QFileIconProvider
from PySide6.QtCore import QFileInfo, QSize, Qt
from ..core import BehaviorDeclarative, AbstractIcon


class Icon(BehaviorDeclarative):
    def __init__(self, 
                 icon: Union[str,
                             QIcon,
                             QIconEngine,
                             QPixmap,
                             QImage,
                             AbstractIcon],
                 color: QColor = None,
                 size: Union[QSize, tuple, int] = None,
                 **kwargs) -> QIcon:

        if isinstance(icon, AbstractIcon):
            icon = icon.value

        if (isinstance(icon, str)
            and icon.endswith(".svg")
            and color is not None):
            return self._colored_icon(icon, color, size)

        if color is not None or size is not None:
            return self._colored_icon(icon, color, size)

        return QIcon(icon)

    @classmethod
    def _colored_icon(cls,
                      icon,
                      color: QColor = None,
                      size: Union[QSize, int] = None) -> QIcon:

        if isinstance(size, int):
            size = QSize(size, size)

        if isinstance(size, (tuple, list)):
            size = QSize(*size[:2])

        icon = QIcon(icon)

        _size = (icon.availableSizes()[0] 
                 if len(icon.availableSizes()) > 0 
                 else icon.actualSize(QSize(32, 32)))

        pixmap = icon.pixmap(size if size is not None else _size)

        pixmap_painter = QPainter(pixmap)
        pixmap_painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_SourceIn)

        if color is not None:
            pixmap_painter.fillRect(pixmap.rect(),
                                    color if color is not None else 0)

        pixmap_painter.end()

        return QIcon(pixmap)

    @classmethod
    def make_svg(cls, 
                 svg: str,
                 color: QColor = None) -> QIcon:

        image = QImage()
        image.fill(Qt.GlobalColor.transparent)
        image.loadFromData(bytes(svg, "utf-8"))
        pixmap = QPixmap.fromImage(image, Qt.ImageConversionFlag.NoFormatConversion)

        if color is not None:
            pixmap_painter = QPainter(pixmap)
            pixmap_painter.setCompositionMode(QPainter.CompositionMode.CompositionMode_SourceIn)
            pixmap_painter.fillRect(pixmap.rect(), color if color is not None else -1)
            pixmap_painter.end()

        return QIcon(pixmap)

    @classmethod
    def file_provider(cls, icon: str | QFileInfo) -> QIcon:
        '''
        param: icon = [file.ext, /path/to/file.ext]
        '''
        return QFileIconProvider().icon(QFileInfo(icon) 
                                        if isinstance(icon, str) 
                                        else icon)


from enum import IntEnum


class TeachingTipTailPositions(IntEnum):
    top = 0
    bottom = 1
    left = 2
    right = 3

    top_left = 4
    top_right = 5

    bottom_left = 6
    bottom_right = 7

    left_top = 8
    left_bottom = 9

    right_top = 10
    right_bottom = 11

    none = 12


from PySide6.QtCore import QPoint, QObject, QPointF
from PySide6.QtGui import QPainterPath, QCursor, QPolygonF
from PySide6.QtWidgets import QWidget, QApplication
from ...enums import TeachingTipTailPositions


class TeachingTipManager(QObject):
    def __init__(self, radius: int = 6):
        super().__init__()
        self.radius = radius

    def doLayout(self, tip: QWidget):
        tip._layout.setContentsMargins(0, 0, 0, 0)

    def position(self, tip: QWidget) -> QPoint:
        pos = self._pos(tip)
        x, y = pos.x(), pos.y()

        rect = QApplication.screenAt(QCursor.pos()).availableGeometry()
        x = min(max(-2, x) if QCursor().pos().x() >=
                0 else x, rect.width() - tip.width() - 4)
        y = min(max(-2, y), rect.height() - tip.height() - 4)

        return QPoint(x, y)

    def _pos(self, tip: QWidget):
        return tip.pos()

    @staticmethod
    def make(position: TeachingTipTailPositions):
        managers = {
            TeachingTipTailPositions.top: TopTailTeachingTipManager,
            TeachingTipTailPositions.bottom: BottomTailTeachingTipManager,
            TeachingTipTailPositions.left: LeftTailTeachingTipManager,
            TeachingTipTailPositions.right: RightTailTeachingTipManager,
            TeachingTipTailPositions.top_right: TopRightTailTeachingTipManager,
            TeachingTipTailPositions.bottom_right: BottomRightTailTeachingTipManager,
            TeachingTipTailPositions.top_left: TopLeftTailTeachingTipManager,
            TeachingTipTailPositions.bottom_left: BottomLeftTailTeachingTipManager,
            TeachingTipTailPositions.left_top: LeftTopTailTeachingTipManager,
            TeachingTipTailPositions.left_bottom: LeftBottomTailTeachingTipManager,
            TeachingTipTailPositions.right_top: RightTopTailTeachingTipManager,
            TeachingTipTailPositions.right_bottom: RightBottomTailTeachingTipManager,
            TeachingTipTailPositions.none: TeachingTipManager,
        }

        if position not in managers:
            raise ValueError(
                f'`{position}` is an invalid teaching tip position.')

        return managers[position]()



class TopTailTeachingTipManager(TeachingTipManager):
    """ Top tail teaching tip manager """

    def doLayout(self, tip):
        tip._layout.setContentsMargins(0, 8, 0, 0)

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pt = tip._layout.contentsMargins().top()

        path = QPainterPath()
        path.addRoundedRect(1, pt, w - 2, h - pt - 1, self.radius, self.radius)
        path.addPolygon(
            QPolygonF([QPointF(w/2 - 7, pt), 
                       QPointF(w/2, 1), 
                       QPointF(w/2 + 7, pt)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QWidget):
        target = tip.target
        pos = target.mapToGlobal(QPoint(0, target.height()))
        x = pos.x() + target.width()//2 - tip.sizeHint().width()//2
        y = pos.y() - tip.layout().contentsMargins().top()
        return QPoint(x, y)


class BottomTailTeachingTipManager(TeachingTipManager):
    """ Bottom tail teaching tip manager """

    def doLayout(self, tip):
        tip._layout.setContentsMargins(0, 0, 0, 8)

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pb = tip._layout.contentsMargins().bottom()

        path = QPainterPath()
        path.addRoundedRect(1, 1, w - 2, h - pb - 1, self.radius, self.radius)
        path.addPolygon(
            QPolygonF([QPointF(w/2 - 7, h - pb), 
                       QPointF(w/2, h - 1), 
                       QPointF(w/2 + 7, h - pb)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QWidget):
        target = tip.target
        pos = target.mapToGlobal(QPoint())
        x = pos.x() + target.width()//2 - tip.sizeHint().width()//2
        y = pos.y() - tip.sizeHint().height() + tip.layout().contentsMargins().bottom()
        return QPoint(x, y)


class LeftTailTeachingTipManager(TeachingTipManager):
    """ Left tail teaching tip manager """

    def doLayout(self, tip):
        tip._layout.setContentsMargins(8, 0, 0, 0)

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pl = 8

        path = QPainterPath()
        path.addRoundedRect(pl, 1, w - pl - 2, h - 2, self.radius, self.radius)
        path.addPolygon(
            QPolygonF([QPointF(pl, h/2 - 7), 
                       QPointF(1, h/2), 
                       QPointF(pl, h/2 + 7)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QWidget):
        target = tip.target
        m = tip.layout().contentsMargins()
        pos = target.mapToGlobal(QPoint(target.width(), 0))
        x = pos.x() - m.left()
        y = pos.y() - tip.view.sizeHint().height()//2 + target.height()//2 - m.top()
        return QPoint(x, y)


class RightTailTeachingTipManager(TeachingTipManager):
    """ Left tail teaching tip manager """

    def doLayout(self, tip):
        tip._layout.setContentsMargins(0, 0, 8, 0)

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pr = 8

        path = QPainterPath()
        path.addRoundedRect(1, 1, w - pr - 1, h - 2, self.radius, self.radius)
        path.addPolygon(
            QPolygonF([QPointF(w - pr, h/2 - 7), 
                       QPointF(w - 1, h/2), 
                       QPointF(w - pr, h/2 + 7)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QWidget):
        target = tip.target
        m = tip.layout().contentsMargins()
        pos = target.mapToGlobal(QPoint(0, 0))
        x = pos.x() - tip.sizeHint().width() + m.right()
        y = pos.y() - tip.view.sizeHint().height()//2 + target.height()//2 - m.top()
        return QPoint(x, y)


class TopLeftTailTeachingTipManager(TopTailTeachingTipManager):
    """ Top left tail teaching tip manager """

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pt = tip._layout.contentsMargins().top()

        path = QPainterPath()
        path.addRoundedRect(1, pt, w - 2, h - pt - 1, self.radius, self.radius)
        path.addPolygon(
            QPolygonF([QPointF(20, pt), QPointF(27, 1), QPointF(34, pt)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QWidget):
        target = tip.target
        pos = target.mapToGlobal(QPoint(0, target.height()))
        x = pos.x() - tip.layout().contentsMargins().left()
        y = pos.y() - tip.layout().contentsMargins().top()
        return QPoint(x, y)


class TopRightTailTeachingTipManager(TopTailTeachingTipManager):
    """ Top right tail teaching tip manager """

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pt = tip._layout.contentsMargins().top()

        path = QPainterPath()
        path.addRoundedRect(1, pt, w - 2, h - pt - 1, self.radius, self.radius)
        path.addPolygon(
            QPolygonF([QPointF(w - 20, pt), QPointF(w - 27, 1), QPointF(w - 34, pt)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QWidget):
        target = tip.target
        pos = target.mapToGlobal(QPoint(target.width(), target.height()))
        x = pos.x() - tip.sizeHint().width() + tip.layout().contentsMargins().left()
        y = pos.y() - tip.layout().contentsMargins().top()
        return QPoint(x, y)


class BottomLeftTailTeachingTipManager(BottomTailTeachingTipManager):
    """ Bottom left tail teaching tip manager """

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pb = tip._layout.contentsMargins().bottom()

        path = QPainterPath()
        path.addRoundedRect(1, 1, w - 2, h - pb - 1, self.radius, self.radius)
        path.addPolygon(
            QPolygonF([QPointF(20, h - pb), QPointF(27, h - 1), QPointF(34, h - pb)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QWidget):
        target = tip.target
        pos = target.mapToGlobal(QPoint())
        x = pos.x() - tip.layout().contentsMargins().left()
        y = pos.y() - tip.sizeHint().height() + tip.layout().contentsMargins().bottom()
        return QPoint(x, y)


class BottomRightTailTeachingTipManager(BottomTailTeachingTipManager):
    """ Bottom right tail teaching tip manager """

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pb = tip._layout.contentsMargins().bottom()

        path = QPainterPath()
        path.addRoundedRect(1, 1, w - 2, h - pb - 1, self.radius, self.radius)
        path.addPolygon(
            QPolygonF([QPointF(w - 20, h - pb), 
                       QPointF(w - 27, h - 1), 
                       QPointF(w - 34, h - pb)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QWidget):
        target = tip.target
        pos = target.mapToGlobal(QPoint(target.width(), 0))
        x = pos.x() - tip.sizeHint().width() + tip.layout().contentsMargins().left()
        y = pos.y() - tip.sizeHint().height() + tip.layout().contentsMargins().bottom()
        return QPoint(x, y)


class LeftTopTailTeachingTipManager(LeftTailTeachingTipManager):
    """ Left top tail teaching tip manager """

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pl = 8

        path = QPainterPath()
        path.addRoundedRect(pl, 1, w - pl - 2, h - 2, self.radius, self.radius)
        path.addPolygon(
            QPolygonF([QPointF(pl, 10), QPointF(1, 17), QPointF(pl, 24)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QWidget):
        target = tip.target
        m = tip.layout().contentsMargins()
        pos = target.mapToGlobal(QPoint(target.width(), 0))
        x = pos.x() - m.left()
        y = pos.y() - m.top()
        return QPoint(x, y)


class LeftBottomTailTeachingTipManager(LeftTailTeachingTipManager):
    """ Left bottom tail teaching tip manager """

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pl = 9

        path = QPainterPath()
        path.addRoundedRect(pl, 1, w - pl - 1, h - 2, self.radius, self.radius)
        path.addPolygon(
            QPolygonF([QPointF(pl, h - 10), QPointF(1, h - 17), QPointF(pl, h - 24)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QWidget):
        target = tip.target
        m = tip.layout().contentsMargins()
        pos = target.mapToGlobal(QPoint(target.width(), target.height()))
        x = pos.x() - m.left()
        y = pos.y() - tip.sizeHint().height() + m.bottom()
        return QPoint(x, y)


class RightTopTailTeachingTipManager(RightTailTeachingTipManager):
    """ Right top tail teaching tip manager """

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pr = 8

        path = QPainterPath()
        path.addRoundedRect(1, 1, w - pr - 1, h - 2, self.radius, self.radius)
        path.addPolygon(
            QPolygonF([QPointF(w - pr, 10), 
                       QPointF(w - 1, 17), 
                       QPointF(w - pr, 24)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QWidget):
        target = tip.target
        m = tip.layout().contentsMargins()
        pos = target.mapToGlobal(QPoint(0, 0))
        x = pos.x() - tip.sizeHint().width() + m.right()
        y = pos.y() - m.top()
        return QPoint(x, y)


class RightBottomTailTeachingTipManager(RightTailTeachingTipManager):
    """ Right bottom tail teaching tip manager """

    def draw(self, tip, painter):
        w, h = tip.width(), tip.height()
        pr = 8

        path = QPainterPath()
        path.addRoundedRect(1, 1, w - pr - 1, h - 2, self.radius, self.radius)
        path.addPolygon(
            QPolygonF([QPointF(w - pr, h-10), 
                       QPointF(w - 1, h-17), 
                       QPointF(w - pr, h-24)]))

        painter.drawPath(path.simplified())

    def _pos(self, tip: QWidget):
        target = tip.target
        m = tip.layout().contentsMargins()
        pos = target.mapToGlobal(QPoint(0, target.height()))
        x = pos.x() - tip.sizeHint().width() + m.right()
        y = pos.y() - tip.sizeHint().height() + m.bottom()
        return QPoint(x, y)


from PySide6.QtWidgets import QGraphicsBlurEffect, QWidget
from ..core import QObjectBase


class BlurEffect(QObjectBase, QGraphicsBlurEffect):
    def __init__(self,  **kwargs) -> QWidget:
        QGraphicsBlurEffect.__init__(self)
        super().__init__(**kwargs)



class Pop:
    # https://github.com/linuxdeepin/dtkwidget/blob/master/src/widgets/dwaterprogress.cpp#L36

    def __init__(self, size, xs, ys, xo=0, yo=0):
        self.size = size
        self.x_speed = xs
        self.y_speed = ys
        self.x_offset = xo
        self.y_offset = yo



#!/usr/bin/python3

from PySide6.QtGui import QPaintEvent
from PySide6.QtWidgets import QWidget
from .base import BehaviorDeclarative


class AbstractPainter(BehaviorDeclarative):
	def __init__(self, child: QWidget) -> QWidget:
		self._parent = child
		self._parent.paintEvent = self._paint

	def update(self):
		self._repaint()

	def _super_paintEvent(self, event: QPaintEvent):
		return self._parent.__class__.paintEvent(self._parent, event)

	def _repaint(self):
		self._parent.update()

	def _paint(self, event: QPaintEvent):
		raise NotImplementedError


class PainterBase(AbstractPainter):
    pass


from PySide6.QtCore import (
    QByteArray, 
    QEasingCurve, 
    QPropertyAnimation, 
    QAbstractAnimation
)

from PySide6.QtWidgets import QWidget
from typing import Any, Sequence, Tuple
from ..core import QObjectDeclarative


class PropertyAnimation(QObjectDeclarative, QPropertyAnimation):
    def __init__(self,
                 *,
                 child: QWidget,
                 property_name: QByteArray | bytes,
                 duration: int = None,
                 start_value: Any = None,
                 end_value: Any = None,
                 easing_curve: QEasingCurve | QEasingCurve.Type = None,
                 loop_count: int = None,
                 key_value: Tuple[float, Any] | Sequence[Tuple[float, Any]] = None,
                 direction: QAbstractAnimation.Direction = None,
                 running: bool = False,
                 **kwargs) -> QWidget:
        QPropertyAnimation.__init__(self)
        super().__init__(**kwargs)

        self.setTargetObject(child)

        if direction is not None:
            self.setDirection(direction)

        if property_name is not None:
            self.setPropertyName(property_name)

        if duration is not None:
            self.setDuration(duration)

        if end_value is not None:
            self.setEndValue(end_value)

        if start_value is not None:
            self.setStartValue(start_value)

        if easing_curve is not None:
            self.setEasingCurve(easing_curve)

        if loop_count is not None:
            self.setLoopCount(loop_count)

        if key_value is not None:
            if not isinstance(key_value[0], int | float):
                self.setKeyValues([(k/100 if 1 < k <= 100 else k, v) 
                                   for k, v in key_value])
            else:
                self.setKeyValueAt(key_value[0]/100
                                   if 1 < key_value[0] <= 100 
                                   else key_value[0],
                                   key_value[-1])
        if running:
            self.start()

        return child


from PySide6.QtWidgets import QWidget, QLayout
from PySide6.QtCore import Qt


class VLayoutItemWrapper:
    def __init__(self, 
                 child: QWidget | QLayout, 
                 stretch: int = None,
                 alignment: Qt.AlignmentFlag = None) -> None:

        if not child:
            raise AttributeError("child should be QWidget | QLayout")

        self.child = child
        self.stretch = stretch
        self.alignment = alignment

    def _yield_attr(self):
        for attr in (
            self.child,
            self.stretch,
            self.alignment
        ):
            if attr is not None:
                yield attr



# built-in
from .grid import GridLayout, GridLayoutItemWrapper, GridAlignment
from .form import FormLayout, FormLayoutItemWrapper
from .horizontal import HLayout, HLayoutItemWrapper
from .vertical import VLayout, VLayoutItemWrapper
from .stacked import StackedLayout

# external
from .expand_layout import ExpandLayout
from .flow_layout import FlowLayout


from .base import (
    AbstractBase,
    NoneCheck,
    DuplicateKeyError,
    BehaviorDeclarative,
    TrackingDeclarative
)

from .init import (
    AndroidInit,
    IOSInit,
    LinuxInit,
    MacOSInit,
    WindowsInit
)

from .api import Api
from .qstyle_sheet import QStyleSheet
from .qobject_base import QObjectBase, QObjectDeclarative
from .widget_base import WidgetBase, WidgetDeclarative
from .tool_base import AbstractTool, ToolBase
from .icons_base import AbstractIcon, IconBase
from .painter_base import AbstractPainter, PainterBase


from typing import Union
from PySide6.QtWidgets import QMenu, QSystemTrayIcon
from PySide6.QtGui import QIcon, QPixmap
from ..core import QObjectBase


class SystemTray(QObjectBase, QSystemTrayIcon):
    def __init__(self, 
                 *,
                 tip: str = None,
                 menu: QMenu = None,
                 **kwargs):
        QSystemTrayIcon.__init__(self)
        super().__init__(**kwargs)

        if menu is not None:
            self.setContextMenu(menu)

        if tip is not None:
            self.setToolTip(tip)

    def alert(self, 
              title: str,
              msg: str = "", 
              icon: Union[QIcon, QPixmap, QSystemTrayIcon.MessageIcon] = QSystemTrayIcon.MessageIcon.Information,
              timeout: int = 1500) -> None:
        self.showMessage(title, msg, icon, timeout)

    def notify(self, 
               title: str,
               msg: str = "", 
               icon: Union[QIcon, QPixmap, QSystemTrayIcon.MessageIcon] = QSystemTrayIcon.MessageIcon.Information,
               timeout: int = 1500) -> None:
        self.alert(title, msg, icon, timeout)


import base64
import hashlib
import os

try:
    from cryptography.fernet import Fernet
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
except ImportError:
    raise Exception('Install "cryptography" Python package to use Flet security utils.')


def __generate_fernet_key(secret_key: str) -> bytes:
    key = hashlib.sha256(secret_key.encode()).digest()
    return base64.urlsafe_b64encode(key)


def __generate_fernet_key_kdf(secret_key: str, salt: bytes) -> bytes:
    kdf = PBKDF2HMAC(algorithm=hashes.SHA256(), length=32, salt=salt, iterations=100000)
    return base64.urlsafe_b64encode(kdf.derive(secret_key.encode("utf-8")))


def encrypt(plain_text: str, secret_key: str) -> str:
    salt = os.urandom(16)
    key = __generate_fernet_key_kdf(secret_key, salt)
    f = Fernet(key)
    return base64.urlsafe_b64encode(
        salt + f.encrypt(plain_text.encode("utf-8"))
    ).decode()


def decrypt(encrypted_data: str, secret_key: str) -> str:
    encrypted_data_bytes = base64.urlsafe_b64decode(encrypted_data)
    salt = encrypted_data_bytes[:16]
    key = __generate_fernet_key_kdf(secret_key, salt)
    f = Fernet(key)
    return f.decrypt(encrypted_data_bytes[16:]).decode("utf-8")



from typing import Union
from PySide6.QtGui import QColor, QFont, QIcon, QPaintEvent, QPainter, QPen
from PySide6.QtCore import QRect, QRectF, Qt
from PySide6.QtWidgets import QWidget
from ..core import WidgetBase


class _ProgressIndicator(QWidget):
    def __init__(self,
                 maximum: int,
                 current: int = 1,
                 step_size: int = 35,
                 font_name: str = "Monospace",
                 steps_content: list[tuple[Union[str, QIcon]]] = None,
                 parent: QWidget | None = None) -> None:
        QWidget.__init__(self, parent)

        self._steps_content = steps_content
        self._current_step = current
        self._step_size = step_size
        self._max_step = maximum
        self._font_name = font_name

        self._style_color = QColor("#5AA9FF")
        self._line_color = QColor(Qt.GlobalColor.gray)
        self._text_color = QColor(Qt.GlobalColor.white)
        self._pen_width = 5

        self.setMinimumHeight(70)

    @property
    def current_step(self) -> int:
        return self._current_step

    @property
    def maximum_step(self) -> int:
        return self._max_step

    @property
    def _is_steps(self) -> bool:
        return self._steps_content is not None

    def setPenWidth(self, width: int):
        self._pen_width = width
        self.update()

    def setFontName(self, font: str):
        self._font_name = font
        self.update()

    def setStyleColor(self, color: QColor = QColor("#5AA9FF")):
        self._style_color = color
        self.update()

    def setLineColor(self, color: QColor = QColor(Qt.GlobalColor.gray)):
        self._line_color = color
        self.update()

    def setTextColor(self, color: QColor = QColor(Qt.GlobalColor.white)):
        self._text_color = color
        self.update()

    def setStepSize(self, size: int):
        self._step_size = size
        self.update()

    def setMaximumSteps(self, steps: int):
        if steps < 2:
            raise ValueError("invalid steps count!")

        self._max_step = steps
        self.update()

    def setCurrentStep(self, step: int):
        if (step > self.maximum_step or step < 1):
            return

        self._current_step = step
        self.update()

    def setNext(self):
        self.setCurrentStep(self.current_step + 1)

    def setPrev(self):
        self.setCurrentStep(self.current_step - 1)

    def _is_icon_step_value(self, value: object):
        return isinstance(value, QIcon)

    def _get_step_value(self, step: int):
        value = 0
        if step == self.current_step:
            value = 1
        elif step < self.current_step:
            value = 2

        return self._steps_content[step - 1][value]

    def paintEvent(self, event: QPaintEvent) -> None:
        qp = QPainter(self)
        qp.setRenderHint(
            QPainter.RenderHint.Antialiasing 
            | QPainter.RenderHint.TextAntialiasing
        )

        left = self.rect().left()
        line_width = 5

        for step in range(1, self._max_step + 1):
            left += ((self.width() / self._max_step)
                     if step > 1
                     else (self.width() / self._max_step / 2))

            if step == self._current_step:
                step_rect = QRectF(left,
                                self.height() // 3 - 5,
                                self._step_size + 10,
                                self._step_size + 10)

            else:
                step_rect = QRectF(left,
                                self.height() // 3,
                                self._step_size,
                                self._step_size)

            # draw rounded step
            if (self._is_steps and not 
                self._is_icon_step_value(self._get_step_value(step))):
                if step < self.current_step or step == self.current_step:
                    qp.setPen(QColor(Qt.GlobalColor.transparent))
                    qp.setBrush(self._style_color)
                else:
                    qp.setPen(QPen(self._line_color, self._pen_width))
                    qp.setBrush(QColor(Qt.GlobalColor.transparent))
            else:
                color = (self._style_color 
                         if step < self.current_step 
                         or step == self.current_step 
                         else self._line_color)
                qp.setPen(QPen(color, self._pen_width))
                qp.setBrush(QColor(Qt.GlobalColor.transparent))

            qp.drawEllipse(step_rect)

            if (not self._is_steps 
                or not self._is_icon_step_value(self._get_step_value(step))):
                # draw step number
                qp.setPen(self._text_color
                        if step == self.current_step 
                        or step < self.current_step
                        else self._style_color)
                font = QFont(self._font_name, 
                            (12 
                            if step == self.current_step 
                            else 11))
                font.setBold(step == self.current_step)
                qp.setFont(font)
                qp.drawText(step_rect,
                            Qt.AlignmentFlag.AlignCenter,
                            self._get_step_value(step)
                            if self._is_steps 
                            else str(step))
            else:
                # draw step icon
                qp.setPen(QColor(Qt.GlobalColor.transparent))
                icon_rect = QRect(
                    step_rect.left()   + (5 // 2) + .5,
                    step_rect.top()    + (5 // 2),
                    step_rect.width()  - 5,
                    step_rect.height() - 5
                )

                if step == self.current_step:
                    icon_rect = step_rect
                elif step < self.current_step:
                    icon_rect = QRect(
                        step_rect.left()   + 5,
                        step_rect.top()    + 5,
                        step_rect.width()  - 10,
                        step_rect.height() - 10
                    )

                mode = (QIcon.Mode.Active 
                        if step == self.current_step 
                        or step < self.current_step 
                        else QIcon.Mode.Disabled)

                self._get_step_value(step).paint(qp,
                                                 icon_rect,
                                                 Qt.AlignmentFlag.AlignCenter,
                                                 mode)

            # draw line
            if step < self._max_step:
                qp.setPen(QColor(Qt.GlobalColor.transparent))

                line_left = left + self._step_size
                if step == 1:
                    line_left = left + self._step_size

                if step == self._current_step:
                    line_rect = QRect(
                        line_left + 12,
                        self.height() // 3 + (self._step_size // 2),
                        (self.width() // self._max_step) - self._step_size - 10,
                        line_width
                    )
                else:
                    line_rect = QRect(
                        line_left,
                        self.height() // 3 + (self._step_size // 2),
                        (self.width() // self._max_step) - self._step_size,
                        line_width
                    )

                if step < self._current_step:
                    qp.setBrush(self._style_color)
                else:
                    qp.setBrush(QColor(Qt.GlobalColor.gray))

                qp.drawRect(line_rect)

        # return super().paintEvent(event)


class ProgressIndicator(WidgetBase, _ProgressIndicator):
    '''
    ProgressIndicator(
        maximum=3
        steps_content = [
            (content, current_content, previous_content),
            (content, current_content, previous_content),
            (content, current_content, previous_content)
        ]
    )
    '''

    def __init__(self,
                 *,
                 maximum: int,
                 current: int = 1,
                 step_size: int = 35,
                 font_name: str = "Monospace",
                 steps_content: list[tuple[Union[str, QIcon]]] = None,
                 style_color: QColor = None,
                 line_color: QColor = None,
                 text_color: QColor = None,
                 pen_width: int = None,
                 **kwargs):
        _ProgressIndicator.__init__(self, maximum, current, step_size, font_name, steps_content)
        super().__init__(**kwargs)

        if style_color is not None:
            self._style_color = style_color
        if line_color is not None:
            self._line_color = line_color
        if text_color is not None:
            self._text_color = text_color
        if pen_width is not None:
            self._pen_width = pen_width


from typing import TypeAlias, Union
from PySide6.QtCore import QObject
from PySide6 import QtWidgets
from enum import Enum


class Widgets(Enum):
    any: QObject = QObject
    abstract_button: QObject = QtWidgets.QAbstractButton
    abstract_graphics_shape_item: QObject = QtWidgets.QAbstractGraphicsShapeItem
    abstract_item_delegate: QObject = QtWidgets.QAbstractItemDelegate
    abstract_item_view: QObject = QtWidgets.QAbstractItemView
    abstract_scroll_area: QObject = QtWidgets.QAbstractScrollArea
    abstract_slider: QObject = QtWidgets.QAbstractSlider
    abstract_spin_box: QObject = QtWidgets.QAbstractSpinBox
    accessible_widget: QObject = QtWidgets.QAccessibleWidget
    application: QObject = QtWidgets.QApplication
    box_layout: QObject = QtWidgets.QBoxLayout
    button_group: QObject = QtWidgets.QButtonGroup
    calendar_widget: QObject = QtWidgets.QCalendarWidget
    check_box: QObject = QtWidgets.QCheckBox
    color_dialog: QObject = QtWidgets.QColorDialog
    colormap: QObject = QtWidgets.QColormap
    column_view: QObject = QtWidgets.QColumnView
    combo_box: QObject = QtWidgets.QComboBox
    command_link_button: QObject = QtWidgets.QCommandLinkButton
    common_style: QObject = QtWidgets.QCommonStyle
    completer: QObject = QtWidgets.QCompleter
    data_widget_mapper: QObject = QtWidgets.QDataWidgetMapper
    date_edit: QObject = QtWidgets.QDateEdit
    date_time_edit: QObject = QtWidgets.QDateTimeEdit
    dial: QObject = QtWidgets.QDial
    dialog: QObject = QtWidgets.QDialog
    dialog_button_box: QObject = QtWidgets.QDialogButtonBox
    dock_widget: QObject = QtWidgets.QDockWidget
    double_spin_box: QObject = QtWidgets.QDoubleSpinBox
    error_message: QObject = QtWidgets.QErrorMessage
    file_dialog: QObject = QtWidgets.QFileDialog
    file_icon_provider: QObject = QtWidgets.QFileIconProvider
    file_system_model: QObject = QtWidgets.QFileSystemModel
    focus_frame: QObject = QtWidgets.QFocusFrame
    font_combo_box: QObject = QtWidgets.QFontComboBox
    font_dialog: QObject = QtWidgets.QFontDialog
    form_layout: QObject = QtWidgets.QFormLayout
    frame: QObject = QtWidgets.QFrame
    gesture: QObject = QtWidgets.QGesture
    gesture_event: QObject = QtWidgets.QGestureEvent
    gesture_recognizer: QObject = QtWidgets.QGestureRecognizer
    graphics_anchor: QObject = QtWidgets.QGraphicsAnchor
    graphics_anchor_layout: QObject = QtWidgets.QGraphicsAnchorLayout
    graphics_blur_effect: QObject = QtWidgets.QGraphicsBlurEffect
    graphics_colorize_effect: QObject = QtWidgets.QGraphicsColorizeEffect
    graphics_drop_shadow_effect: QObject = QtWidgets.QGraphicsDropShadowEffect
    graphics_effect: QObject = QtWidgets.QGraphicsEffect
    graphics_ellipse_item: QObject = QtWidgets.QGraphicsEllipseItem
    graphics_grid_layout: QObject = QtWidgets.QGraphicsGridLayout
    graphics_item: QObject = QtWidgets.QGraphicsItem
    graphics_item_animation: QObject = QtWidgets.QGraphicsItemAnimation
    graphics_item_group: QObject = QtWidgets.QGraphicsItemGroup
    graphics_layout: QObject = QtWidgets.QGraphicsLayout
    graphics_layout_item: QObject = QtWidgets.QGraphicsLayoutItem
    graphics_line_item: QObject = QtWidgets.QGraphicsLineItem
    graphics_linear_layout: QObject = QtWidgets.QGraphicsLinearLayout
    graphics_object: QObject = QtWidgets.QGraphicsObject
    graphics_opacity_effect: QObject = QtWidgets.QGraphicsOpacityEffect
    graphics_path_item: QObject = QtWidgets.QGraphicsPathItem
    graphics_pixmap_item: QObject = QtWidgets.QGraphicsPixmapItem
    graphics_polygon_item: QObject = QtWidgets.QGraphicsPolygonItem
    graphics_proxy_widget: QObject = QtWidgets.QGraphicsProxyWidget
    graphics_rect_item: QObject = QtWidgets.QGraphicsRectItem
    graphics_rotation: QObject = QtWidgets.QGraphicsRotation
    graphics_scale: QObject = QtWidgets.QGraphicsScale
    graphics_scene: QObject = QtWidgets.QGraphicsScene
    graphics_scene_context_menu_event: QObject = QtWidgets.QGraphicsSceneContextMenuEvent
    graphics_scene_drag_drop_event: QObject = QtWidgets.QGraphicsSceneDragDropEvent
    graphics_scene_event: QObject = QtWidgets.QGraphicsSceneEvent
    graphics_scene_help_event: QObject = QtWidgets.QGraphicsSceneHelpEvent
    graphics_scene_hover_event: QObject = QtWidgets.QGraphicsSceneHoverEvent
    graphics_scene_mouse_event: QObject = QtWidgets.QGraphicsSceneMouseEvent
    graphics_scene_move_event: QObject = QtWidgets.QGraphicsSceneMoveEvent
    graphics_scene_resize_event: QObject = QtWidgets.QGraphicsSceneResizeEvent
    graphics_scene_wheel_event: QObject = QtWidgets.QGraphicsSceneWheelEvent
    graphics_simple_text_item: QObject = QtWidgets.QGraphicsSimpleTextItem
    graphics_text_item: QObject = QtWidgets.QGraphicsTextItem
    graphics_transform: QObject = QtWidgets.QGraphicsTransform
    graphics_view: QObject = QtWidgets.QGraphicsView
    graphics_widget: QObject = QtWidgets.QGraphicsWidget
    grid_layout: QObject = QtWidgets.QGridLayout
    group_box: QObject = QtWidgets.QGroupBox
    hbox_layout: QObject = QtWidgets.QHBoxLayout
    header_view: QObject = QtWidgets.QHeaderView
    input_dialog: QObject = QtWidgets.QInputDialog
    int_list: QObject = QtWidgets.QIntList
    item_delegate: QObject = QtWidgets.QItemDelegate
    item_editor_creator_base: QObject = QtWidgets.QItemEditorCreatorBase
    item_editor_factory: QObject = QtWidgets.QItemEditorFactory
    key_sequence_edit: QObject = QtWidgets.QKeySequenceEdit
    lcdnumber: QObject = QtWidgets.QLCDNumber
    label: QObject = QtWidgets.QLabel
    layout: QObject = QtWidgets.QLayout
    layout_item: QObject = QtWidgets.QLayoutItem
    line_edit: QObject = QtWidgets.QLineEdit
    list_view: QObject = QtWidgets.QListView
    list_widget: QObject = QtWidgets.QListWidget
    list_widget_item: QObject = QtWidgets.QListWidgetItem
    main_window: QObject = QtWidgets.QMainWindow
    mdi_area: QObject = QtWidgets.QMdiArea
    mdi_sub_window: QObject = QtWidgets.QMdiSubWindow
    menu: QObject = QtWidgets.QMenu
    menu_bar: QObject = QtWidgets.QMenuBar
    message_box: QObject = QtWidgets.QMessageBox
    pan_gesture: QObject = QtWidgets.QPanGesture
    pinch_gesture: QObject = QtWidgets.QPinchGesture
    plain_text_document_layout: QObject = QtWidgets.QPlainTextDocumentLayout
    plain_text_edit: QObject = QtWidgets.QPlainTextEdit
    point_flist: QObject = QtWidgets.QPointFList
    point_list: QObject = QtWidgets.QPointList
    progress_bar: QObject = QtWidgets.QProgressBar
    progress_dialog: QObject = QtWidgets.QProgressDialog
    proxy_style: QObject = QtWidgets.QProxyStyle
    push_button: QObject = QtWidgets.QPushButton
    radio_button: QObject = QtWidgets.QRadioButton
    rubber_band: QObject = QtWidgets.QRubberBand
    scroll_area: QObject = QtWidgets.QScrollArea
    scroll_bar: QObject = QtWidgets.QScrollBar
    scroller: QObject = QtWidgets.QScroller
    scroller_properties: QObject = QtWidgets.QScrollerProperties
    size_grip: QObject = QtWidgets.QSizeGrip
    size_policy: QObject = QtWidgets.QSizePolicy
    slider: QObject = QtWidgets.QSlider
    spacer_item: QObject = QtWidgets.QSpacerItem
    spin_box: QObject = QtWidgets.QSpinBox
    splash_screen: QObject = QtWidgets.QSplashScreen
    splitter: QObject = QtWidgets.QSplitter
    splitter_handle: QObject = QtWidgets.QSplitterHandle
    stacked_layout: QObject = QtWidgets.QStackedLayout
    stacked_widget: QObject = QtWidgets.QStackedWidget
    status_bar: QObject = QtWidgets.QStatusBar
    style: QObject = QtWidgets.QStyle
    style_factory: QObject = QtWidgets.QStyleFactory
    style_hint_return: QObject = QtWidgets.QStyleHintReturn
    style_hint_return_mask: QObject = QtWidgets.QStyleHintReturnMask
    style_hint_return_variant: QObject = QtWidgets.QStyleHintReturnVariant
    style_option: QObject = QtWidgets.QStyleOption
    style_option_button: QObject = QtWidgets.QStyleOptionButton
    style_option_combo_box: QObject = QtWidgets.QStyleOptionComboBox
    style_option_complex: QObject = QtWidgets.QStyleOptionComplex
    style_option_dock_widget: QObject = QtWidgets.QStyleOptionDockWidget
    style_option_focus_rect: QObject = QtWidgets.QStyleOptionFocusRect
    style_option_frame: QObject = QtWidgets.QStyleOptionFrame
    style_option_graphics_item: QObject = QtWidgets.QStyleOptionGraphicsItem
    style_option_group_box: QObject = QtWidgets.QStyleOptionGroupBox
    style_option_header: QObject = QtWidgets.QStyleOptionHeader
    style_option_header_v2: QObject = QtWidgets.QStyleOptionHeaderV2
    style_option_menu_item: QObject = QtWidgets.QStyleOptionMenuItem
    style_option_progress_bar: QObject = QtWidgets.QStyleOptionProgressBar
    style_option_rubber_band: QObject = QtWidgets.QStyleOptionRubberBand
    style_option_size_grip: QObject = QtWidgets.QStyleOptionSizeGrip
    style_option_slider: QObject = QtWidgets.QStyleOptionSlider
    style_option_spin_box: QObject = QtWidgets.QStyleOptionSpinBox
    style_option_tab: QObject = QtWidgets.QStyleOptionTab
    style_option_tab_bar_base: QObject = QtWidgets.QStyleOptionTabBarBase
    style_option_tab_widget_frame: QObject = QtWidgets.QStyleOptionTabWidgetFrame
    style_option_title_bar: QObject = QtWidgets.QStyleOptionTitleBar
    style_option_tool_bar: QObject = QtWidgets.QStyleOptionToolBar
    style_option_tool_box: QObject = QtWidgets.QStyleOptionToolBox
    style_option_tool_button: QObject = QtWidgets.QStyleOptionToolButton
    style_option_view_item: QObject = QtWidgets.QStyleOptionViewItem
    style_painter: QObject = QtWidgets.QStylePainter
    styled_item_delegate: QObject = QtWidgets.QStyledItemDelegate
    swipe_gesture: QObject = QtWidgets.QSwipeGesture
    system_tray_icon: QObject = QtWidgets.QSystemTrayIcon
    tab_bar: QObject = QtWidgets.QTabBar
    tab_widget: QObject = QtWidgets.QTabWidget
    table_view: QObject = QtWidgets.QTableView
    table_widget: QObject = QtWidgets.QTableWidget
    table_widget_item: QObject = QtWidgets.QTableWidgetItem
    table_widget_selection_range: QObject = QtWidgets.QTableWidgetSelectionRange
    tap_and_hold_gesture: QObject = QtWidgets.QTapAndHoldGesture
    tap_gesture: QObject = QtWidgets.QTapGesture
    text_browser: QObject = QtWidgets.QTextBrowser
    text_edit: QObject = QtWidgets.QTextEdit
    tile_rules: QObject = QtWidgets.QTileRules
    time_edit: QObject = QtWidgets.QTimeEdit
    tool_bar: QObject = QtWidgets.QToolBar
    tool_box: QObject = QtWidgets.QToolBox
    tool_button: QObject = QtWidgets.QToolButton
    tool_tip: QObject = QtWidgets.QToolTip
    tree_view: QObject = QtWidgets.QTreeView
    tree_widget: QObject = QtWidgets.QTreeWidget
    tree_widget_item: QObject = QtWidgets.QTreeWidgetItem
    tree_widget_item_iterator: QObject = QtWidgets.QTreeWidgetItemIterator
    undo_view: QObject = QtWidgets.QUndoView
    vbox_layout: QObject = QtWidgets.QVBoxLayout
    whats_this: QObject = QtWidgets.QWhatsThis
    widget: QObject = QtWidgets.QWidget
    widget_action: QObject = QtWidgets.QWidgetAction
    widget_item: QObject = QtWidgets.QWidgetItem
    wizard: QObject = QtWidgets.QWizard
    wizard_page: QObject = QtWidgets.QWizardPage


WidgetTypeVar: TypeAlias = Union[QObject, Widgets]



from PySide6.QtWidgets import QWidget, QLayout
from typing import Union


class FormLayoutItemWrapper:
    def __init__(
        self,
        label: Union[QWidget, str],
        field: Union[QWidget, QLayout]
    ) -> None:

        if not label:
            raise AttributeError("label should be QWidget | str")

        if not field:
            raise AttributeError("field should be QWidget | QLayout")

        self.label = label
        self.field = field

    # def __init__(self, row: Union[QWidget, QLayout]):
    #     if not row:
    #         raise AttributeError("row should be QWidget | QLayout")
    #     self.row = row

    def _yield_attr(self):
        # if hasattr(self, "row"):
        #     yield self.row
        # else:
            for attr in (
                self.label,
                self.field
            ):
                if attr is not None:
                    yield attr


#!/usr/bin/python3

from PySide6.QtWidgets import QStackedWidget, QWidget


class Routes:
    def __init__(self, 
                 stacked_widget: QStackedWidget = None,
                 index: str = "/") -> None:

        self._stacked_widget = stacked_widget
        self._index = index
        self._routes: dict = {}

    def add(self, route: str, widget: QWidget) -> None:
        self._routes[route] = self._stacked_widget.addWidget(widget)

    def push(self, route: str) -> None:
        self._stacked_widget.setCurrentIndex(self._routes[route])

    def pop(self) -> None:
        self._stacked_widget.setCurrentIndex(self._routes[self._index])

    def remove(self, route: str) -> None:
        index = self._routes.pop(route)
        self._stacked_widget.removeWidget(self.widget(index))

    def current_route(self) -> str:
        return {v: k for k, v in self._routes.items()}.get(
            self._stacked_widget.currentIndex()
        )

    def stackedWidget(self) -> QStackedWidget:
        return self._stacked_widget



#!/usr/bin/python3

from typing import Any, Union, Sequence
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QGraphicsEffect
from .base import BehaviorDeclarative, AbstractBase
from .qstyle_sheet import QStyleSheet
from ..enums.signals import SignalTypeVar
from ..enums.events import EventTypeVar


class WidgetBase(AbstractBase):
    def __init__(self,
                 *,
                 uid: str = None, 
                 signals: SignalTypeVar = None, 
                 events: EventTypeVar = None,
                 methods: Sequence[tuple[str, Any]] = None,
                 qss: Union[str, dict, QStyleSheet] = None,
                 attrs: Union[list[Qt.WidgetAttribute], 
                              dict[Qt.WidgetAttribute, bool]] = None,
                 flags: Union[list[Qt.WindowType], 
                              dict[Qt.WindowType, bool]] = None,
                 effect: QGraphicsEffect = None,
                 **kwargs):
        super().__init__(
            uid=uid, 
            signals=signals, 
            events=events, 
            methods=methods, 
            **kwargs)

        self.setUpdatesEnabled(True)
        self.setMouseTracking(True)

        self._set_qss(
            qss if qss is not None
            else QStyleSheet(qss)
        )

        if effect is not None:
            self._set_effect(effect)

        if attrs is not None:
            self._set_attrs(attrs)

        if flags is not None:
            self._set_flags(flags)

    def _set_effect(self, effect: QGraphicsEffect):
        effect.setParent(self)
        self.setGraphicsEffect(effect)

    def _set_flags(self, flags):
        if isinstance(flags, dict):
            for flag, on in flags.items():
                self.setWindowFlag(flag, on)
        else:
            for flag in flags:
                self.setWindowFlag(flag)

    def _set_attrs(self, attrs: dict):
        if isinstance(attrs, dict):
            for attr, on in attrs.items():
                self.setAttribute(attr, on)
        else:
            for attr in attrs:
                self.setAttribute(attr)

    def _set_qss(self, qss: Union[str, dict, QStyleSheet]):
        self.qss = (QStyleSheet(qss)
                    if isinstance(qss, (str, dict))
                    else qss)

        self.qss._set_parent(self)
        self.qss._set_qss(self.qss._qss)


class WidgetDeclarative(WidgetBase, BehaviorDeclarative):
    pass



from typing import Union, List, Tuple, Callable, Any, TypeAlias 
from PySide6 import QtCore
from enum import Enum, auto


class Events(Enum):
    action_event: QtCore.QEvent = auto()
    change_event: QtCore.QEvent = auto()
    child_event: QtCore.QEvent = auto()
    close_event: QtCore.QEvent = auto()
    context_menu_event: QtCore.QEvent = auto()
    custom_event: QtCore.QEvent = auto()
    drag_enter_event: QtCore.QEvent = auto()
    drag_leave_event: QtCore.QEvent = auto()
    drag_move_event: QtCore.QEvent = auto()
    drop_event: QtCore.QEvent = auto()
    editor_event: QtCore.QEvent = auto()
    enter_event: QtCore.QEvent = auto()
    focus_in_event: QtCore.QEvent = auto()
    focus_out_event: QtCore.QEvent = auto()
    grab_keyboard_event: QtCore.QEvent = auto()
    grab_mouse_event: QtCore.QEvent = auto()
    help_event: QtCore.QEvent = auto()
    hide_event: QtCore.QEvent = auto()
    hover_enter_event: QtCore.QEvent = auto()
    hover_leave_event: QtCore.QEvent = auto()
    hover_move_event: QtCore.QEvent = auto()
    input_method_event: QtCore.QEvent = auto()
    is_input_event: QtCore.QEvent = auto()
    is_pointer_event: QtCore.QEvent = auto()
    is_single_point_event: QtCore.QEvent = auto()
    key_press_event: QtCore.QEvent = auto()
    key_release_event: QtCore.QEvent = auto()
    leave_event: QtCore.QEvent = auto()
    mouse_double_click_event: QtCore.QEvent = auto()
    mouse_move_event: QtCore.QEvent = auto()
    mouse_press_event: QtCore.QEvent = auto()
    mouse_release_event: QtCore.QEvent = auto()
    move_event: QtCore.QEvent = auto()
    native_event: QtCore.QEvent = auto()
    paint_event: QtCore.QEvent = auto()
    polish_event: QtCore.QEvent = auto()
    resend_prepare_event: QtCore.QEvent = auto()
    resize_event: QtCore.QEvent = auto()
    scene_event: QtCore.QEvent = auto()
    send_event: QtCore.QEvent = auto()
    show_event: QtCore.QEvent = auto()
    tablet_event: QtCore.QEvent = auto()
    timer_event: QtCore.QEvent = auto()
    ungrab_keyboard_event: QtCore.QEvent = auto()
    ungrab_mouse_event: QtCore.QEvent = auto()
    viewport_event: QtCore.QEvent = auto()
    wheel_event: QtCore.QEvent = auto()
    widget_event: QtCore.QEvent = auto()
    window_frame_event: QtCore.QEvent = auto()

EventTypeVar: TypeAlias = List[Tuple[Union[Events, str], Callable[..., Any]]]



# coding:utf-8
from copy import deepcopy



def exceptionHandler(*default):
    """ decorator for exception handling

    Parameters
    ----------
    *default:
        the default value returned when an exception occurs
    """

    def outer(func):

        def inner(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except BaseException:
                value = deepcopy(default)
                if len(value) == 0:
                    return None
                elif len(value) == 1:
                    return value[0]

                return value

        return inner

    return outer



#!/usr/bin/python3

import math


class Vector2D(complex):
    """
    Simple immutable 2D vector class based on the Python complex number type

    Create and access - coordinates

    >>> v = Vector(1, 2)
    >>> v.x, v.y
    (1.0, 2.0)

    Create and access - angle and magnitude (length)

    >>> v = Vector.polar(math.pi, 2)
    >>> v
    Vector(-2.0, 0.0)
    >>> v.magnitude  # Length of the vector, alias for abs(v)
    2.0
    >>> v.radians
    3.141592653589793
    >>> v.degrees
    180.0

    Arithmetic operations

    >>> Vector(1, 1) + 2
    Vector(3.0, 1.0)
    >>> Vector(0.1, 0.1) + Vector(0.2, 0.2)  == Vector(0.3, 0.3)  # Float tolerance 10 decimals
    True
    >>> Vector(2, 3) - Vector(1, 1)
    Vector(1.0, 2.0)
    >>> Vector(1, 1) * 2
    Vector(2.0, 2.0)
    >>> round(Vector.polar(math.pi / 4, 1), 1)
    Vector(0.7, 0.7)

    Get a new vector by adjusting one of the coordinates
    >>> v = Vector()
    >>> v.with_x(1)
    Vector(1.0, 0.0)
    >>> v.with_y(2)
    Vector(0.0, 2.0)

    Get a new vector by adjusting angle or magnitude

    >>> v = Vector(1, 2)
    >>> v = v.with_magnitude(4.47213595499958)  # Twice as long
    >>> v.x, v.y
    (2.0, 4.0)

    >>> v = Vector.polar(math.pi, 2)
    >>> v
    Vector(-2.0, 0.0)
    >>> v.with_radians(0)
    Vector(2.0, 0.0)
    >>> v.with_degrees(90)
    Vector(0.0, 2.0)
    """

    abs_tol = 1e-10

    x = complex.real
    y = complex.imag
    __add__ = lambda self, other: type(self)(complex.__add__(self, other))
    __sub__ = lambda self, other: type(self)(complex.__sub__(self, other))
    __mul__ = lambda self, other: type(self)(complex.__mul__(self, other))
    __truediv__ = lambda self, other: type(self)(complex.__truediv__(self, other))
    __len__ = lambda self: 2
    __round__ = lambda self, ndigits=None: type(self)(
        round(self.x, ndigits), round(self.y, ndigits)
    )

    def __eq__(self, other):
        return math.isclose(self.x, other.x, abs_tol=self.abs_tol) and math.isclose(
            self.y, other.y, abs_tol=self.abs_tol
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __iter__(self):
        return iter([self.x, self.y])

    def __str__(self):
        return str(tuple(self))

    def __repr__(self):
        return f"{type(self).__name__}{str(self)}"

    @classmethod
    def polar(cls, radians, magnitude):
        return cls(
            round(math.cos(radians) * magnitude, 10),
            round(math.sin(radians) * magnitude, 10),
        )

    @property
    def magnitude(self):
        return abs(self)

    @property
    def degrees(self):
        return math.degrees(self.radians)

    @property
    def radians(self):
        return math.atan2(self.y, self.x)

    def with_x(self, value):
        return type(self)(value, self.y)

    def with_y(self, value):
        return type(self)(self.x, value)

    def with_magnitude(self, value):
        return self * value / abs(self)

    def with_radians(self, value):
        magnitude = abs(self)
        return type(self).polar(value, magnitude)

    def with_degrees(self, value):
        radians = math.radians(value)
        magnitude = abs(self)
        return type(self).polar(radians, magnitude)


def boundary(value, minvalue, maxvalue):
    '''Limit a value between a minvalue and maxvalue.'''
    return min(max(value, minvalue), maxvalue)


def intersection(set1, set2):
    '''Return the intersection of 2 lists.'''
    return [s for s in set1 if s in set2]


def difference(set1, set2):
    '''Return the difference between 2 lists.'''
    return [s for s in set1 if s not in set2]


def interpolate(value_from, value_to, step=10):
    '''Interpolate between two values. This can be useful for smoothing some
    transitions. For example::

        # instead of setting directly
        self.pos = pos

        # use interpolate, and you'll have a nicer transition
        self.pos = interpolate(self.pos, new_pos)

    .. warning::
        These interpolations work only on lists/tuples/doubles with the same
        dimensions. No test is done to check the dimensions are the same.
    '''
    if type(value_from) in (list, tuple):
        out = []
        for x, y in zip(value_from, value_to):
            out.append(interpolate(x, y, step))
        return out
    else:
        return value_from + (value_to - value_from) / float(step)

def format_bytes_to_human(size, precision=2):
    '''Format a byte value to a human readable representation (B, KB, MB...).

    .. versionadded:: 1.0.8

    :Parameters:
        `size`: int
            Number that represents the bytes value
        `precision`: int, defaults to 2
            Precision after the comma

    Examples::

        >>> format_bytes_to_human(6463)
        '6.31 KB'
        >>> format_bytes_to_human(646368746541)
        '601.98 GB'

    '''
    size = int(size)
    fmt = '%%1.%df %%s' % precision
    for unit in ['B', 'KB', 'MB', 'GB', 'TB']:
        if size < 1024.0:
            return fmt % (size, unit)
        size /= 1024.0


'''
Geometry utilities
==================

This module contains some helper functions for geometric calculations.
'''

__all__ = ('circumcircle', 'minimum_bounding_circle')

from .vector import Vector


def circumcircle(a, b, c):
    '''
    Computes the circumcircle of a triangle defined by a, b, c.
    See: http://en.wikipedia.org/wiki/Circumscribed_circle

    :Parameters:
        `a`: iterable containing at least 2 values (for x and y)
            The 1st point of the triangle.
        `b`: iterable containing at least 2 values (for x and y)
            The 2nd point of the triangle.
        `c`: iterable containing at least 2 values (for x and y)
            The 3rd point of the triangle.

    :Return:
        A tuple that defines the circle :
         * The first element in the returned tuple is the center as (x, y)
         * The second is the radius (float)
    '''
    P = Vector(a[0], a[1])
    Q = Vector(b[0], b[1])
    R = Vector(c[0], c[1])

    mPQ = (P + Q) * .5
    mQR = (Q + R) * .5

    numer = -(- mPQ.y * R.y + mPQ.y * Q.y + mQR.y * R.y - mQR.y * Q.y -
              mPQ.x * R.x + mPQ.x * Q.x + mQR.x * R.x - mQR.x * Q.x)
    denom = (-Q.x * R.y + P.x * R.y - P.x * Q.y +
             Q.y * R.x - P.y * R.x + P.y * Q.x)

    t = numer / denom

    cx = -t * (Q.y - P.y) + mPQ.x
    cy = t * (Q.x - P.x) + mPQ.y

    return ((cx, cy), (P - (cx, cy)).length())


def minimum_bounding_circle(points):
    '''
    Returns the minimum bounding circle for a set of points.

    For a description of the problem being solved, see the `Smallest Circle
    Problem <http://en.wikipedia.org/wiki/Smallest_circle_problem>`_.

    The function uses Applet's Algorithm, the runtime is ``O(h^3, *n)``,
    where h is the number of points in the convex hull of the set of points.
    **But** it runs in linear time in almost all real world cases.
    See: http://tinyurl.com/6e4n5yb

    :Parameters:
        `points`: iterable
            A list of points (2 tuple with x,y coordinates)

    :Return:
        A tuple that defines the circle:
            * The first element in the returned tuple is the center (x, y)
            * The second the radius (float)

    '''
    points = [Vector(p[0], p[1]) for p in points]

    if len(points) == 1:
        return (points[0].x, points[0].y), 0.0

    if len(points) == 2:
        p1, p2 = points
        return (p1 + p2) * .5, ((p1 - p2) * .5).length()

    # determine a point P with the smallest y value
    P = min(points, key=lambda p: p.y)

    # find a point Q such that the angle of the line segment
    # PQ with the x axis is minimal
    def x_axis_angle(q):
        if q == P:
            return 1e10  # max val if the same, to skip
        return abs((q - P).angle((1, 0)))
    Q = min(points, key=x_axis_angle)

    for p in points:
        # find R such that angle PRQ is minimal
        def angle_pq(r):
            if r in (P, Q):
                return 1e10  # max val if the same, to skip
            return abs((r - P).angle(r - Q))
        R = min(points, key=angle_pq)

        # check for case 1 (angle PRQ is obtuse), the circle is determined
        # by two points, P and Q. radius = |(P-Q)/2|, center = (P+Q)/2
        if angle_pq(R) > 90.0:
            return (P + Q) * .5, ((P - Q) * .5).length()

        # if angle RPQ is obtuse, make P = R, and try again
        if abs((R - P).angle(Q - P)) > 90:
            P = R
            continue

        # if angle PQR is obtuse, make Q = R, and try again
        if abs((P - Q).angle(R - Q)) > 90:
            Q = R
            continue

        # all angles were acute..we just need the circle through the
        # two points furthest apart!
        break

    # find the circumcenter for triangle given by P,Q,R
    return circumcircle(P, Q, R)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

import math

from PySide6.QtCore import Slot, QTimer, QSizeF, Qt, QRectF, QPointF, QRect, QPoint, QSize
from PySide6.QtGui import QImage, QColor, QPainter, QLinearGradient, QGradient, QPainterPath, QPixmap, QBrush, QPen
from PySide6.QtWidgets import QProgressBar, QGraphicsDropShadowEffect
from PySide6.QtSvg import QSvgRenderer
from ..tools._widgets.water_progress_bar import Pop
from ..core import WidgetBase


_WATER_FRONT = """
<svg xmlns="http://www.w3.org/2000/svg" width="383" height="115" viewBox="0 0 383 115">
  <path fill="#01C4FF" 
        fill-rule="evenodd" 
        d="M383,115 L383,14.1688789 C380.269872,14.0716143 377.092672,13.5814974 373.063461,12.4722672 C368.696509,11.2699114 362.241136,10.1727531 357.649256,10.1227411 C347.007291,10.0071963 342.744795,10.6014761 332.930121,12.0276784 C326.157898,13.0120512 317.51313,12.4953762 311.375303,10.33762 C305.58601,8.30230681 299.587109,8.09191178 293.164466,8.16675723 C284.09108,8.27264456 276.303198,11.8021073 267.219716,11.3406179 C260.695053,11.0091595 256.565913,8.56512814 248.546835,8.86450991 C241.871757,9.11387975 235.569934,13.1896798 228.881972,13.3297132 C219.538394,13.525622 215.498041,10.7384053 208.282229,8.42337018 C201.688974,6.30769299 190.725982,6.45048568 185.454442,8.65549452 C170.142255,15.0597811 162.05946,9.31703167 150.536236,5.36712375 C147.826999,4.43862637 144.672431,3.20971247 141.663406,2.90998579 C135.153716,2.26155522 129.812539,3.9788615 123.613779,5.46231888 C115.747555,7.3451819 106.643181,6.73503633 99.4869089,3.84572629 C96.4124243,2.60474055 93.6255416,0.951587506 90.1882469,0.261077932 C79.652131,-1.85528907 69.7970674,9.59778831 58.8051757,9.35186757 C49.4744806,9.14319709 42.6942497,2.4740197 33.3934986,1.93078665 C20.5224457,1.17888312 19.3845731,15.343297 0,13.8463882 L0,115 L383,115 Z"
    />
</svg>
"""

_WATER_BACK = """
<svg xmlns="http://www.w3.org/2000/svg" width="383" height="115" viewBox="0 0 383 115">
  <path fill="#007DFF" 
        fill-rule="evenodd" 
        d="M383,115 L383,14.1688789 C380.269872,14.0716143 377.092672,13.5814974 373.063461,12.4722672 C368.696509,11.2699114 362.241136,10.1727531 357.649256,10.1227411 C347.007291,10.0071963 342.744795,10.6014761 332.930121,12.0276784 C326.157898,13.0120512 317.51313,12.4953762 311.375303,10.33762 C305.58601,8.30230681 299.587109,8.09191178 293.164466,8.16675723 C284.09108,8.27264456 276.303198,11.8021073 267.219716,11.3406179 C260.695053,11.0091595 256.565913,8.56512814 248.546835,8.86450991 C241.871757,9.11387975 235.569934,13.1896798 228.881972,13.3297132 C219.538394,13.525622 215.498041,10.7384053 208.282229,8.42337018 C201.688974,6.30769299 190.725982,6.45048568 185.454442,8.65549452 C170.142255,15.0597811 162.05946,9.31703167 150.536236,5.36712375 C147.826999,4.43862637 144.672431,3.20971247 141.663406,2.90998579 C135.153716,2.26155522 129.812539,3.9788615 123.613779,5.46231888 C115.747555,7.3451819 106.643181,6.73503633 99.4869089,3.84572629 C96.4124243,2.60474055 93.6255416,0.951587506 90.1882469,0.261077932 C79.652131,-1.85528907 69.7970674,9.59778831 58.8051757,9.35186757 C49.4744806,9.14319709 42.6942497,2.4740197 33.3934986,1.93078665 C20.5224457,1.17888312 19.3845731,15.343297 0,13.8463882 L0,115 L383,115 Z"
    />
</svg>
"""



class _WaterProgressBar(QProgressBar):
    def __init__(self, 
                 interval: int = 30,
                 text_visible: bool = True,
                ):
        super(_WaterProgressBar, self).__init__()

        self.waterFrontImage = QImage()
        self.waterBackImage = QImage()
        self.waterFrontSvg = QSvgRenderer(_WATER_FRONT.encode())
        self.waterBackSvg = QSvgRenderer(_WATER_BACK.encode())

        self.interval = interval
        self.pops: list = [Pop(7, -1.8, 0.6),
                           Pop(8, 1.2, 1.0),
                           Pop(11, 0.8, 1.6)]

        self.setTextVisible(text_visible)
        self.resizePixmap(self.size())

        self.timer = QTimer(self)
        self.timer.setInterval(self.interval)
        self.timer.timeout.connect(self.on_timeout)
        self.timer.start()

        self.frontXOffset = self.width()
        self.backXOffset = 0

        effect = QGraphicsDropShadowEffect(self)
        effect.setOffset(0, 6)
        effect.setColor(QColor(1, 153, 248, 255 * 5 / 20))
        effect.setBlurRadius(12)
        self.setGraphicsEffect(effect)

    def set_interval(self, interval: int):
        self.interval = interval
        self.timer.setInterval(interval)

    @Slot()
    def start(self):
        self.timer.start()

    @Slot()
    def stop(self):
        self.timer.stop()

    def resizePixmap(self, sz):
        # https://github.com/linuxdeepin/dtkwidget/blob/master/src/widgets/dwaterprogress.cpp#L192
        # resize water
        waterWidth = 500 * sz.width() / 100
        waterHeight = 110 * sz.height() / 100
        waterSize = QSizeF(waterWidth, waterHeight).toSize()

        if self.waterFrontImage.size() != waterSize:
            image = QImage(waterWidth, waterHeight, QImage.Format.Format_ARGB32)
            image.fill(Qt.GlobalColor.transparent)

            waterPainter = QPainter(image)
            self.waterFrontSvg.render(waterPainter)
            self.waterFrontImage = image

        if self.waterBackImage.size() != waterSize:
            image = QImage(waterWidth, waterHeight, QImage.Format.Format_ARGB32)
            image.fill(Qt.GlobalColor.transparent)  # partly transparent red-ish background
            
            waterPainter = QPainter(image)
            self.waterBackSvg.render(waterPainter)
            self.waterBackImage = image

    def on_timeout(self):
        # interval can not be zero, and limit to 1
        self.interval = max(1, self.interval)

        # move 60% per second
        frontXDeta = 40.0 / (1000.0 / self.interval)

        # move 90% per second
        backXDeta = 60.0 / (1000.0 / self.interval)

        canvasWidth = int(self.width() * self.devicePixelRatioF())
        self.frontXOffset -= frontXDeta * canvasWidth / 100
        self.backXOffset += backXDeta * canvasWidth / 100

        if self.frontXOffset > canvasWidth:
            self.frontXOffset = canvasWidth

        if self.frontXOffset < - (self.waterFrontImage.width() - canvasWidth):
            self.frontXOffset = canvasWidth

        if self.backXOffset > self.waterBackImage.width():
            self.backXOffset = 0

        # update pop
        # move 25% per second default
        speed = 25 / (1000.0 / self.interval)  # 100 / self.height()
        for pop in self.pops:
            # yOffset 0 ~ 100
            pop.y_offset += speed * pop.y_speed
            if pop.y_offset < 0:
                pass

            if pop.y_offset > self.value():
                pop.y_offset = 0

            pop.x_offset = math.sin((pop.y_offset / 100) * 2 * 3.14) * 18 * pop.x_speed + 50

        self.update()

    def paint(self, painter):
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        pixelRatio = self.devicePixelRatioF()
        rect = QRectF(0, 0, self.width() * pixelRatio, self.height() * pixelRatio)
        sz = QSizeF(self.width() * pixelRatio, self.height() * pixelRatio).toSize()

        self.resizePixmap(sz)

        yOffset = rect.toRect().topLeft().y() + (100 - self.value() - 10) * sz.height() / 100

        # draw water
        waterImage = QImage(sz, QImage.Format.Format_ARGB32_Premultiplied)
        waterPainter = QPainter()
        waterPainter.begin(waterImage)
        waterPainter.setRenderHint(QPainter.RenderHint.Antialiasing)
        waterPainter.setCompositionMode(QPainter.CompositionMode.CompositionMode_Source)

        pointStart = QPointF(sz.width() / 2, 0)
        pointEnd = QPointF(sz.width() / 2, sz.height())
        linear = QLinearGradient(pointStart, pointEnd)
        startColor = QColor('#1F08FF')
        startColor.setAlphaF(1)
        endColor = QColor('#50FFF7')
        endColor.setAlphaF(0.28)
        linear.setColorAt(0, startColor)
        linear.setColorAt(1, endColor)
        linear.setSpread(QGradient.Spread.PadSpread)
        waterPainter.setPen(Qt.PenStyle.NoPen)
        waterPainter.setBrush(linear)
        waterPainter.drawEllipse(waterImage.rect().center(), 
                                 sz.width() / 2 + 1, 
                                 sz.height() / 2 + 1)

        waterPainter.setCompositionMode(QPainter.CompositionMode.CompositionMode_SourceOver)
        waterPainter.drawImage(int(self.backXOffset), yOffset, self.waterBackImage)
        waterPainter.drawImage(int(self.backXOffset) - self.waterBackImage.width(), yOffset,
                               self.waterBackImage)
        waterPainter.drawImage(int(self.frontXOffset), yOffset, self.waterFrontImage)
        waterPainter.drawImage(int(self.frontXOffset) - self.waterFrontImage.width(), yOffset,
                               self.waterFrontImage)

        # draw pop
        if self.value() > 30:
            for pop in self.pops:
                popPath = QPainterPath()
                popPath.addEllipse(pop.x_offset * sz.width() / 100, (100 - pop.y_offset) * sz.height() / 100,
                                   pop.size * sz.width() / 100, pop.size * sz.height() / 100)
                waterPainter.fillPath(popPath, QColor(255, 255, 255, 255 * 0.3))

        if self.isTextVisible():
            font = waterPainter.font()
            rectValue = QRect()
            progressText = self.text().strip('%')

            if progressText == '100':
                font.setPixelSize(sz.height() * 35 / 100)
                waterPainter.setFont(font)

                rectValue.setWidth(sz.width() * 60 / 100)
                rectValue.setHeight(sz.height() * 35 / 100)
                rectValue.moveCenter(rect.center().toPoint())
                waterPainter.setPen(Qt.GlobalColor.white)
                waterPainter.drawText(rectValue, Qt.AlignmentFlag.AlignCenter, progressText)
            else:
                font.setPixelSize(sz.height() * 40 / 100)
                waterPainter.setFont(font)

                rectValue.setWidth(sz.width() * 45 / 100)
                rectValue.setHeight(sz.height() * 40 / 100)
                rectValue.moveCenter(rect.center().toPoint())
                rectValue.moveLeft(rect.left() + rect.width() * 0.45 * 0.5)

                waterPainter.setPen(Qt.GlobalColor.white)
                waterPainter.drawText(rectValue, Qt.AlignmentFlag.AlignCenter, progressText)
                font.setPixelSize(font.pixelSize() / 2)
                waterPainter.setFont(font)
                rectPerent = QRect(QPoint(rectValue.right(), rectValue.bottom() - rect.height() * 20 / 100),
                                   QPoint(rectValue.right() + rect.width() * 20 / 100, rectValue.bottom()))

                waterPainter.drawText(rectPerent, Qt.AlignmentFlag.AlignCenter, '%')

        waterPainter.end()

        maskPixmap = QPixmap(sz)
        maskPixmap.fill(Qt.GlobalColor.transparent)
        path = QPainterPath()
        path.addEllipse(QRectF(0, 0, sz.width(), sz.height()))
        maskPainter = QPainter()
        maskPainter.begin(maskPixmap)
        maskPainter.setRenderHint(QPainter.RenderHint.Antialiasing)
        maskPainter.setPen(QPen(Qt.GlobalColor.white, 1))
        maskPainter.fillPath(path, QBrush(Qt.GlobalColor.white))
        maskPainter.end()

        mode = QPainter.CompositionMode.CompositionMode_SourceIn
        contentImage = QImage(sz, QImage.Format.Format_ARGB32_Premultiplied)
        contentPainter = QPainter()
        contentPainter.begin(contentImage)
        contentPainter.setCompositionMode(QPainter.CompositionMode.CompositionMode_Source)
        contentPainter.fillRect(contentImage.rect(), Qt.GlobalColor.transparent)
        contentPainter.setCompositionMode(QPainter.CompositionMode.CompositionMode_SourceOver)
        contentPainter.drawImage(0, 0, maskPixmap.toImage())
        contentPainter.setCompositionMode(mode)
        contentPainter.drawImage(0, 0, waterImage)
        contentPainter.setCompositionMode(QPainter.CompositionMode.CompositionMode_DestinationOver)
        contentPainter.end()

        contentImage.setDevicePixelRatio(pixelRatio)
        painter.drawImage(self.rect(), contentImage)

    def paintEvent(self, event):
        painter = QPainter(self)
        self.paint(painter)

    def sizeHint(self):
        return QSize(100, 100)


class WaterProgressBar(WidgetBase, _WaterProgressBar):
    def __init__(self, 
                 *,
                 interval: int = 30,
                 text_visible: bool = True,
                 **kwargs):
        _WaterProgressBar.__init__(self, interval, text_visible)
        super().__init__(**kwargs)


from PySide6.QtWidgets import QFrame, QGraphicsProxyWidget, QGraphicsScene, QGraphicsView, QWidget
from PySide6.QtGui import QPalette, QPainter
from PySide6.QtCore import Qt
from ..core import WidgetBase


class _GraphicsView(QGraphicsView):
    def __init__(self, child: QWidget):
        QGraphicsView.__init__(self)

        self._graphics_scene = QGraphicsScene(self)

        self.setScene(self._graphics_scene)

        self.setBackgroundBrush(Qt.GlobalColor.transparent)
        self.setWindowFlags(Qt.WindowType.FramelessWindowHint)
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)

        self.setFrameShape(QFrame.Shape.NoFrame)
        self.setFrameShadow(QFrame.Shadow.Raised)

        self.setBackgroundRole(QPalette.ColorRole.NoRole)
        self.setStyleSheet("background: transparent")
        self.setContentsMargins(0, 0, 0, 0)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        self.setInteractive(True)
        self.setUpdatesEnabled(True)

        self.setViewportUpdateMode(QGraphicsView.ViewportUpdateMode.SmartViewportUpdate)

        self.setRenderHints(
            QPainter.RenderHint.Antialiasing
            | QPainter.RenderHint.TextAntialiasing
            | QPainter.RenderHint.SmoothPixmapTransform
        )

        self._proxy = self.scene().addWidget(child)
        self._proxy.setTransformOriginPoint(self._proxy.boundingRect().center())
        self._proxy.setContentsMargins(0, 0, 0, 0)

        self.centerOn(self.sceneRect().center())

    @property
    def graphics_scene(self):
        return self._graphics_scene

    @property
    def proxy(self) -> QGraphicsProxyWidget:
        return self._proxy

    def update_size(self):
        self.fitInView(self._proxy.graphicsItem(),
                       Qt.AspectRatioMode.KeepAspectRatio)

    def resizeEvent(self, event):
        self.fitInView(self._proxy.graphicsItem(),
                       Qt.AspectRatioMode.KeepAspectRatio)
        super().resizeEvent(event)


class GraphicsView(WidgetBase, _GraphicsView):
    def __init__(self, 
                 *,
                 child: QWidget,
                 **kwargs):
        _GraphicsView.__init__(self, child)
        super().__init__(**kwargs)


from typing import Union
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QGridLayout, QWidget, QLayout, QLayoutItem
from .item_wrapper import GridLayoutItemWrapper, GridAlignment
from ...core import QObjectBase


class GridLayout(QObjectBase, QGridLayout):
    def __init__(self,
                 *,
                 children: list[Union[QWidget, QLayoutItem, GridLayoutItemWrapper]] = None,
                 **kwargs):
        QGridLayout.__init__(self)
        super().__init__(**kwargs)

        self._set_children(children)

    def _set_children(self, children: list[Union[QWidget, 
                                                 GridLayoutItemWrapper]]) -> None:
        if not children:
            return

        for child in children:
            if isinstance(child, QLayoutItem):
                self.addItem(child)
            
            elif isinstance(child, QWidget):
                self.addWidget(child)

            elif isinstance(child, GridAlignment):
                self.addWidget(child._child)
                self.setAlignment(child._child, 
                                  child._alignment)

            elif isinstance(child, GridLayoutItemWrapper):
                child.attrs.update({
                    "row":     child.attrs.get("row", self.rowCount()),
                    "column":  child.attrs.get("column", self.columnCount()),
                    "alignment": child.attrs.get("alignment", Qt.AlignmentFlag.AlignCenter)
                })

                values = child.attrs.values()

                if isinstance(child.child, QWidget):
                    self.addWidget(*values)

                elif isinstance(child.child, QLayoutItem):
                    self.addItem(*values)

                elif isinstance(child.child, QLayout):
                    self.addLayout(*values)

    @staticmethod
    def wrapper(
            child: Union[QWidget, QLayout, QLayoutItem],
            row: int = None,
            column: int = None,
            row_span: int = None,
            column_span: int = None,
            alignment: Qt.AlignmentFlag = None
        ) -> GridLayoutItemWrapper:
        return GridLayoutItemWrapper(
            child,
            row,
            column,
            row_span,
            column_span,
            alignment
        )


from .colors import Colors, ColorTypeVar
from .signals import Signals, SignalTypeVar
from .widgets import Widgets, WidgetTypeVar
from .events import Events, EventTypeVar
from .env_vars import EnvVars
from .qpa_plugins import QPA_Platforms
from .animation import AnimationPath
from .position import Positions
from .size import Sizes
from .smooth_scroll import SmoothMode
from .teaching_tip_tails import TeachingTipTailPositions
from .clipboard import ClipboardDataTypes
from .theme import Theme


from typing import Dict, Union
from PySide6.QtCore import Signal
from PySide6.QtGui import QCloseEvent, QShowEvent
from PySide6.QtWidgets import (
    QMainWindow,
    QSystemTrayIcon, 
    QWidget, 
    QToolBar,
    QStatusBar,
    QDockWidget,
    QStackedWidget,
    QLayout
)
from ...utils._routes import Routes
from ...core import WidgetBase


class RoutingWindow(WidgetBase, QMainWindow):
    startup_changed = Signal()

    def __init__(self, 
                 *,
                 index: str = "/",
                 routes: Dict[str, Union[QWidget, QLayout]] = None,
                 stacked_widget: QStackedWidget = None,
                 tool_bar: QToolBar = None,
                 status_bar: QStatusBar = None,
                 dock_widget: QDockWidget = None,
                 sys_tray: QSystemTrayIcon = None,
                 **kwargs):
        QMainWindow.__init__(self)
        super().__init__(**kwargs)

        self.__is_startup = False
        stacked_widget = (stacked_widget 
                          if stacked_widget is not None 
                          else QStackedWidget())

        stacked_widget.setParent(self)
        self._routes = Routes(stacked_widget, index)

        self.setAnimated(True)
        self.setUpdatesEnabled(True)
        self.setDocumentMode(True)

        if routes is not None:
            self._set_routes(routes)

        if tool_bar is not None:
            self._set_tool_bar(tool_bar)
    
        if status_bar is not None:
            self._set_status_bar(status_bar)

        if dock_widget is not None:
            self._set_dock_widget(dock_widget)

        if sys_tray is not None:
            sys_tray.setParent(self)
            sys_tray.show()

    def closeEvent(self, event: QCloseEvent) -> None:
        super().closeEvent(event)
        self.__is_startup = False

    def showEvent(self, event: QShowEvent) -> None:
        super().showEvent(event)
        if not self.__is_startup:
            self.__is_startup = True
            self.startup_changed.emit()

    def _set_routes(self, routes: Dict[str, QWidget]) -> None:
        for route, child in routes.items():
            if child is not None:
                if isinstance(child, QLayout):
                    widget = QWidget(self)
                    widget.setLayout(child)
                    self._routes.add(route, widget)
                else:
                    self._routes.add(route, child)

        self.setCentralWidget(self._routes.stackedWidget())

    def _set_tool_bar(self, tool_bar: QToolBar) -> None:
        self.addToolBar(tool_bar)

    def _set_status_bar(self, status_bar: QStatusBar) -> None:
        self.setStatusBar(status_bar)

    def _set_dock_widget(self, dock_widget: QDockWidget) -> None:
        self.addDockWidget(dock_widget)

    def push(self, route: str) -> None:
        self.route.push(route)

    @property
    def route(self) -> Routes:
        return self._routes